from typing import List, Optional, Tuple

from rastervision.core.box import Box
from rastervision.pipeline.config import (Config, register_config, Field,
                                          validator, ConfigError)
from rastervision.core.data.raster_transformer import RasterTransformerConfig


def rs_config_upgrader(cfg_dict: dict, version: int) -> dict:
    if version == 6:
        # removed in version 7
        if cfg_dict.get('extent_crop') is not None:
            raise ConfigError('RasterSourceConfig.extent_crop is deprecated.')
        try:
            del cfg_dict['extent_crop']
        except KeyError:
            pass
    return cfg_dict


@register_config('raster_source', upgrader=rs_config_upgrader)
class RasterSourceConfig(Config):
    """Configure a :class:`.RasterSource`."""

    channel_order: Optional[List[int]] = Field(
        None,
        description=
        'The sequence of channel indices to use when reading imagery.')
    transformers: List[RasterTransformerConfig] = []
    extent: Optional[Tuple[int, int, int, int]] = Field(
        None,
        description='Use-specified extent in pixel coords in the form '
        '(ymin, xmin, ymax, xmax). Useful for cropping the raster source so '
        'that only part of the raster is read from.')

    def build(self, tmp_dir, use_transformers=True):
        raise NotImplementedError()

    def update(self, pipeline=None, scene=None):
        for t in self.transformers:
            t.update(pipeline, scene)

    @validator('extent')
    def validate_extent(cls, v):
        if v is None:
            return None
        return Box(*v)



from typing import Any, Optional
from pyproj import Transformer

import rasterio as rio
from rasterio.transform import (rowcol, xy)
from rasterio import Affine

from rastervision.core.data.crs_transformer import (CRSTransformer,
                                                    IdentityCRSTransformer)


class RasterioCRSTransformer(CRSTransformer):
    """Transformer for a RasterioRasterSource."""

    def __init__(self,
                 transform: Affine,
                 image_crs: Any,
                 map_crs: Any = 'epsg:4326',
                 round_pixels: bool = True):
        """Constructor.

        Args:
            transform (Affine): Rasterio affine transform.
            image_crs (Any): CRS of image in format that PyProj can handle
                eg. wkt or init string.
            map_crs (Any): CRS of the labels. Defaults to "epsg:4326".
            round_pixels (bool): If True, round outputs of map_to_pixel and
                inputs of pixel_to_map to integers. Defaults to False.
        """

        if (image_crs is None) or (image_crs == map_crs):
            self.map2image = lambda *args, **kws: args[:2]
            self.image2map = lambda *args, **kws: args[:2]
        else:
            self.map2image = Transformer.from_crs(
                map_crs, image_crs, always_xy=True).transform
            self.image2map = Transformer.from_crs(
                image_crs, map_crs, always_xy=True).transform

        self.round_pixels = round_pixels

        super().__init__(transform, image_crs, map_crs)

    def __repr__(self) -> str:  # pragma: no cover
        cls_name = type(self).__name__

        image_crs_str = str(self.image_crs)
        if len(image_crs_str) > 70:
            image_crs_str = image_crs_str[:70] + '...'

        map_crs_str = str(self.image_crs)
        if len(map_crs_str) > 70:
            map_crs_str = map_crs_str[:70] + '...'

        transform_str = (
            '\n\t\t' + (str(self.transform).replace('\n', '\n\t\t')))
        out = f"""{cls_name}(
            image_crs="{image_crs_str}",
            map_crs="{map_crs_str}",
            round_pixels="{self.round_pixels}",
            transform={transform_str})
        """
        return out

    def _map_to_pixel(self, map_point):
        """Transform point from map to pixel-based coordinates.

        Args:
            map_point: (x, y) tuple in map coordinates

        Returns:
            (x, y) tuple in pixel coordinates
        """
        image_point = self.map2image(*map_point)
        x, y = image_point
        if self.round_pixels:
            row, col = rowcol(self.transform, x, y)
        else:
            row, col = rowcol(self.transform, x, y, op=lambda x: x)
        pixel_point = (col, row)
        return pixel_point

    def _pixel_to_map(self, pixel_point):
        """Transform point from pixel to map-based coordinates.

        Args:
            pixel_point: (x, y) tuple in pixel coordinates

        Returns:
            (x, y) tuple in map coordinates
        """
        col, row = pixel_point
        if self.round_pixels:
            col, row = int(col), int(row)
        image_point = xy(self.transform, row, col, offset='center')
        map_point = self.image2map(*image_point)
        return map_point

    @classmethod
    def from_dataset(cls,
                     dataset,
                     map_crs: Optional[str] = 'epsg:4326',
                     **kwargs) -> 'RasterioCRSTransformer':
        transform = dataset.transform
        image_crs = None if dataset.crs is None else dataset.crs.wkt
        map_crs = image_crs if map_crs is None else map_crs

        no_crs_tf = (image_crs is None) or (image_crs == map_crs)
        no_affine_tf = (transform is None) or (transform == Affine.identity())
        if no_crs_tf and no_affine_tf:
            return IdentityCRSTransformer()

        if transform is None:
            transform = Affine.identity()

        return cls(transform, image_crs, map_crs, **kwargs)

    @classmethod
    def from_uri(cls, uri: str, map_crs: Optional[str] = 'epsg:4326',
                 **kwargs) -> 'RasterioCRSTransformer':
        with rio.open(uri) as ds:
            return cls.from_dataset(ds, map_crs=map_crs, **kwargs)



from typing import TYPE_CHECKING

from rastervision.pipeline.file_system import json_to_file
from rastervision.core.data.label import ChipClassificationLabels
from rastervision.core.data.label_store import LabelStore
from rastervision.core.data.label_store.utils import boxes_to_geojson
from rastervision.core.data.label_source import (
    ChipClassificationLabelSourceConfig)
from rastervision.core.data.vector_source import (GeoJSONVectorSourceConfig)

if TYPE_CHECKING:
    from rastervision.core.data import ClassConfig, CRSTransformer


class ChipClassificationGeoJSONStore(LabelStore):
    """Storage for chip classification predictions."""

    def __init__(self, uri: str, class_config: 'ClassConfig',
                 crs_transformer: 'CRSTransformer'):
        """Constructor.

        Args:
            uri: uri of GeoJSON file containing labels
            class_config: ClassConfig
            crs_transformer: CRSTransformer to convert from map coords in label
                in GeoJSON file to pixel coords.
        """
        self.uri = uri
        self.class_config = class_config
        self.crs_transformer = crs_transformer

    def save(self, labels: ChipClassificationLabels) -> None:
        """Save labels to URI if writable.

        Note that if the grid is inferred from polygons, only the grid will be
        written, not the original polygons.
        """
        boxes = labels.get_cells()
        class_ids = labels.get_class_ids()
        scores = list(labels.get_scores())
        geojson = boxes_to_geojson(
            boxes,
            class_ids,
            self.crs_transformer,
            self.class_config,
            scores=scores)
        json_to_file(geojson, self.uri)

    def get_labels(self) -> ChipClassificationLabels:
        vs = GeoJSONVectorSourceConfig(uri=self.uri)
        ls = ChipClassificationLabelSourceConfig(vector_source=vs).build(
            self.class_config, self.crs_transformer)
        return ls.get_labels()

    def empty_labels(self) -> ChipClassificationLabels:
        return ChipClassificationLabels()



# flake8: noqa

from rastervision.core.data.label_source.label_source import *
from rastervision.core.data.label_source.label_source_config import *
from rastervision.core.data.label_source.chip_classification_label_source import *
from rastervision.core.data.label_source.chip_classification_label_source_config import *
from rastervision.core.data.label_source.semantic_segmentation_label_source import *
from rastervision.core.data.label_source.semantic_segmentation_label_source_config import *
from rastervision.core.data.label_source.object_detection_label_source import *
from rastervision.core.data.label_source.object_detection_label_source_config import *

__all__ = [
    LabelSource.__name__,
    LabelSourceConfig.__name__,
    SemanticSegmentationLabelSource.__name__,
    SemanticSegmentationLabelSourceConfig.__name__,
    ChipClassificationLabelSource.__name__,
    ChipClassificationLabelSourceConfig.__name__,
    ObjectDetectionLabelSource.__name__,
    ObjectDetectionLabelSourceConfig.__name__,
]



from typing import Any
from abc import ABC, abstractmethod, abstractproperty

from rastervision.core.box import Box


class LabelSource(ABC):
    """An interface for storage of labels for a scene.

    An LabelSource is a read source of labels for a scene
    that could be backed by a file, a database, an API, etc. The difference
    between LabelSources and Labels can be understood by analogy to the
    difference between a database and result sets queried from a database.
    """

    @abstractmethod
    def get_labels(self, window=None):
        """Return labels overlapping with window.

        Args:
            window: Box

        Returns:
            Labels overlapping with window. If window is None,
                returns all labels.
        """
        pass

    @abstractproperty
    def extent(self) -> Box:
        pass

    def __getitem__(self, key: Any) -> Any:
        if isinstance(key, Box):
            raise NotImplementedError()
        elif isinstance(key, slice):
            key = [key]
        elif isinstance(key, tuple):
            pass
        else:
            raise TypeError('Unsupported key type.')
        slices = list(key)
        assert 1 <= len(slices) <= 2
        assert all(s is not None for s in slices)
        assert isinstance(slices[0], slice)
        if len(slices) == 1:
            h, = slices
            w = slice(None, None)
        else:
            assert isinstance(slices[1], slice)
            h, w = slices

        if any(x is not None and x < 0
               for x in [h.start, h.stop, w.start, w.stop]):
            raise NotImplementedError()

        ymin, xmin, ymax, xmax = self.extent
        _ymin = 0 if h.start is None else h.start
        _xmin = 0 if w.start is None else w.start
        _ymax = ymax if h.stop is None else h.stop
        _xmax = xmax if w.stop is None else w.stop
        window = Box(_ymin, _xmin, _ymax, _xmax)
        return self[window]



from abc import (ABC, abstractmethod)
from pprint import pformat


class EvaluationItem(ABC):
    @abstractmethod
    def merge(self, other):
        """Merges another item from a different scene into this one.

        This is used to average metrics over scenes. Merges by taking a
        weighted average (by gt_count) of the metrics.
        """
        pass

    @abstractmethod
    def to_json(self) -> dict:
        return self.__dict__

    def __repr__(self):
        return pformat(self.to_json())



import os
from subprocess import Popen

from rastervision.pipeline.file_system import (download_or_copy,
                                               upload_or_copy)

COGIFY = 'COGIFY'

DEFAULT_BLOCK_SIZE = 512
DEFAULT_RESAMPLE_METHOD = 'near'
DEFAULT_COMPRESSION = 'deflate'
DEFAULT_OVERVIEWS = [2, 4, 8, 16, 32]


def gdal_cog_commands(input_path,
                      tmp_dir,
                      block_size=DEFAULT_BLOCK_SIZE,
                      resample_method=DEFAULT_RESAMPLE_METHOD,
                      compression=DEFAULT_COMPRESSION,
                      overviews=None):
    """
    GDAL commands to create a COG from an input file.
    Returns a tuple (commands, output_path)
    """

    if not overviews:
        overviews = DEFAULT_OVERVIEWS

    def get_output_path(command):
        fname = os.path.splitext(os.path.basename(input_path))[0]
        return os.path.join(tmp_dir, '{}-{}.tif'.format(fname, command))

    compression = compression.lower()

    def add_compression(cmd, overview=False):
        if compression != 'none':
            if not overview:
                return cmd[:1] + ['-co', 'compress={}'.format(compression)
                                  ] + cmd[1:]
            else:
                return cmd[:1] + [
                    '--config', 'COMPRESS_OVERVIEW', compression
                ] + cmd[1:]
        else:
            return cmd

    # Step 1: Translate to a GeoTiff.
    translate_path = get_output_path('translate')
    translate = add_compression([
        'gdal_translate', '-of', 'GTiff', '-co', 'tiled=YES', '-co',
        'BIGTIFF=IF_SAFER', input_path, translate_path
    ])

    # Step 2: Add overviews
    add_overviews = add_compression(
        ['gdaladdo', '-r', resample_method, translate_path] + list(
            map(lambda x: str(x), overviews)),
        overview=True)

    # Step 3: Translate to COG
    output_path = get_output_path('cog')

    create_cog = add_compression([
        'gdal_translate', '-co', 'TILED=YES', '-co', 'COPY_SRC_OVERVIEWS=YES',
        '-co', 'BLOCKXSIZE={}'.format(block_size), '-co',
        'BLOCKYSIZE={}'.format(block_size), '-co', 'BIGTIFF=IF_SAFER',
        '--config', 'GDAL_TIFF_OVR_BLOCKSIZE',
        str(block_size), translate_path, output_path
    ])

    return ([translate, add_overviews, create_cog], output_path)


def run_cmd(cmd):
    p = Popen(cmd)
    (out, err) = p.communicate(input)
    if p.returncode != 0:
        s = 'Command failed:\n'
        s += ' '.join(cmd) + '\n\n'
        if out:
            s += out + '\n\n'
        if err:
            s += err
        raise Exception(s)


def create_cog(source_uri,
               dest_uri,
               local_dir,
               block_size=DEFAULT_BLOCK_SIZE,
               resample_method=DEFAULT_RESAMPLE_METHOD,
               compression=DEFAULT_COMPRESSION,
               overviews=None):
    local_path = download_or_copy(source_uri, local_dir)

    commands, output_path = gdal_cog_commands(
        local_path,
        local_dir,
        block_size=block_size,
        resample_method=resample_method,
        compression=compression,
        overviews=overviews)
    for command in commands:
        run_cmd(command)

    upload_or_copy(output_path, dest_uri)



from rastervision.pipeline.config import register_config
from rastervision.core.evaluation.evaluator_config import EvaluatorConfig


@register_config('classification_evaluator')
class ClassificationEvaluatorConfig(EvaluatorConfig):
    """Configure a :class:`.ClassificationEvaluator`."""

    pass



from typing import Iterable, Optional

from rastervision.core.analyzer import Analyzer
from rastervision.core.raster_stats import RasterStats
from rastervision.core.data import Scene


class StatsAnalyzer(Analyzer):
    """Compute imagery statistics of scenes."""

    def __init__(self,
                 stats_uri: Optional[str] = None,
                 sample_prob: float = 0.1):
        self.stats_uri = stats_uri
        self.sample_prob = sample_prob

    def compute_stats(self, scenes: Iterable[Scene]) -> RasterStats:
        stats = RasterStats()
        stats.compute(
            [s.raster_source for s in scenes], sample_prob=self.sample_prob)
        return stats

    def process(self, scenes: Iterable[Scene], tmp_dir: str) -> None:
        stats = self.compute_stats(scenes)
        if self.stats_uri is not None:
            stats.save(self.stats_uri)
        else:
            raise ValueError('Cannot save stats because stats_uri is not set.')



from typing import TYPE_CHECKING, Dict, Tuple
import numpy as np
from shapely.strtree import STRtree

from rastervision.core.evaluation import (ClassificationEvaluation,
                                          ClassEvaluationItem)

if TYPE_CHECKING:
    from shapely.geometry import Polygon
    from rastervision.core.data import ObjectDetectionLabels
    from rastervision.core.data.class_config import ClassConfig


def compute_metrics(
        gt_labels: 'ObjectDetectionLabels',
        pred_labels: 'ObjectDetectionLabels',
        num_classes: int,
        iou_thresh: float = 0.5) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    gt_geoms = [b.to_shapely() for b in gt_labels.get_boxes()]
    gt_classes = gt_labels.get_class_ids()
    pred_geoms = [b.to_shapely() for b in pred_labels.get_boxes()]
    pred_classes = pred_labels.get_class_ids()

    for pred_geom, class_id in zip(pred_geoms, pred_classes):
        pred_geom.class_id = class_id
    pred_tree = STRtree(pred_geoms)

    def iou(a: 'Polygon', b: 'Polygon') -> float:
        return a.intersection(b).area / a.union(b).area

    def is_matched(geom) -> bool:
        return hasattr(geom, 'iou_matched')

    tp = np.zeros((num_classes, ))
    fp = np.zeros((num_classes, ))
    fn = np.zeros((num_classes, ))

    for gt_geom, gt_class in zip(gt_geoms, gt_classes):
        matches = [
            g for g in pred_tree.query(gt_geom)
            if (not is_matched(g)) and (g.class_id == gt_class)
        ]
        ious = np.array([iou(m, gt_geom) for m in matches])
        if (ious > iou_thresh).any():
            max_ind = np.argmax(ious)
            matches[max_ind].iou_matched = True
            tp[gt_class] += 1
        else:
            fn[gt_class] += 1

    for class_id in range(num_classes):
        pred_not_matched = np.array([not is_matched(g) for g in pred_geoms])
        fp[class_id] = np.sum(pred_not_matched[pred_classes == class_id])

    return tp, fp, fn


class ObjectDetectionEvaluation(ClassificationEvaluation):
    def __init__(self, class_config: 'ClassConfig', iou_thresh: float = 0.5):
        super().__init__()
        self.class_config = class_config
        self.iou_thresh = iou_thresh

    def compute(self, ground_truth_labels: 'ObjectDetectionLabels',
                prediction_labels: 'ObjectDetectionLabels'):
        self.class_to_eval_item = ObjectDetectionEvaluation.compute_eval_items(
            ground_truth_labels,
            prediction_labels,
            self.class_config,
            iou_thresh=self.iou_thresh)
        self.compute_avg()

    @staticmethod
    def compute_eval_items(
            gt_labels: 'ObjectDetectionLabels',
            pred_labels: 'ObjectDetectionLabels',
            class_config: 'ClassConfig',
            iou_thresh: float = 0.5) -> Dict[int, ClassEvaluationItem]:
        num_classes = len(class_config)
        tps, fps, fns = compute_metrics(gt_labels, pred_labels, num_classes,
                                        iou_thresh)
        class_to_eval_item = {}
        for class_id, (tp, fp, fn) in enumerate(zip(tps, fps, fns)):
            class_name = class_config.get_name(class_id)
            eval_item = ClassEvaluationItem(
                class_id=class_id, class_name=class_name, tp=tp, fp=fp, fn=fn)
            class_to_eval_item[class_id] = eval_item

        return class_to_eval_item



from typing import List, Optional

import numpy as np

from rastervision.core.data.raster_transformer import RasterTransformer


class MinMaxTransformer(RasterTransformer):
    """Transforms chips by scaling values in each channel to span 0-255."""

    def transform(self,
                  chip: np.ndarray,
                  channel_order: Optional[List[int]] = None) -> np.ndarray:
        _, _, c = chip.shape
        channel_mins = chip.reshape(-1, c).min(axis=0)
        channel_maxs = chip.reshape(-1, c).max(axis=0)
        chip_normalized = (chip - channel_mins) / (channel_maxs - channel_mins)
        chip_normalized = (255 * chip_normalized).astype(np.uint8)
        return chip_normalized



# flake8: noqa
from rastervision.core.data.vector_transformer.vector_transformer import *
from rastervision.core.data.vector_transformer.vector_transformer_config import *
from rastervision.core.data.vector_transformer.class_inference_transformer import *
from rastervision.core.data.vector_transformer.class_inference_transformer_config import *
from rastervision.core.data.vector_transformer.buffer_transformer import *
from rastervision.core.data.vector_transformer.buffer_transformer_config import *
from rastervision.core.data.vector_transformer.shift_transformer import *
from rastervision.core.data.vector_transformer.shift_transformer_config import *

__all__ = [
    VectorTransformer.__name__,
    VectorTransformerConfig.__name__,
    BufferTransformer.__name__,
    BufferTransformerConfig.__name__,
    ClassInferenceTransformer.__name__,
    ClassInferenceTransformerConfig.__name__,
    ShiftTransformer.__name__,
    ShiftTransformerConfig.__name__,
]



from typing import Union

from rastervision.core.data.raster_source import (RasterSourceConfig,
                                                  RasterizedSourceConfig)
from rastervision.core.data.label_source import (
    LabelSourceConfig, SemanticSegmentationLabelSource)
from rastervision.pipeline.config import (register_config, Field)


def ss_label_source_config_upgrader(cfg_dict: dict, version: int) -> dict:
    if version < 4:
        try:
            # removed in version 4
            del cfg_dict['rgb_class_config']
        except KeyError:
            pass
    return cfg_dict


@register_config(
    'semantic_segmentation_label_source',
    upgrader=ss_label_source_config_upgrader)
class SemanticSegmentationLabelSourceConfig(LabelSourceConfig):
    """Configure a :class:`.SemanticSegmentationLabelSource`."""

    raster_source: Union[RasterSourceConfig, RasterizedSourceConfig] = Field(
        ..., description='The labels in the form of rasters.')

    def build(self, class_config, crs_transformer, extent, tmp_dir):
        if isinstance(self.raster_source, RasterizedSourceConfig):
            rs = self.raster_source.build(class_config, crs_transformer,
                                          extent)
        else:
            rs = self.raster_source.build(tmp_dir)
        return SemanticSegmentationLabelSource(rs, class_config)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Numpy BoxList classes and functions."""

from typing import List, Tuple

import numpy as np


class BoxList(object):
    """A list of bounding boxes as a [y_min, x_min, y_max, x_max] numpy array.

    It is assumed that all bounding boxes within a given list correspond to a
    single image. Optionally, users can add additional related fields (such as
    objectness/classification scores).
    """

    def __init__(self, data: np.ndarray):
        """Constructor.

        Args:
            data (np.ndarray): Box coords as a [N, 4] numpy array.

        Raises:
            ValueError: If bbox data is not a numpy array.
            ValueError: If invalid dimensions for bbox data.
        """
        if not isinstance(data, np.ndarray):
            raise ValueError('data must be a numpy array.')
        if len(data.shape) != 2 or data.shape[1] != 4:
            raise ValueError('Invalid dimensions for box data.')
        if data.dtype != np.float32 and data.dtype != np.float64:
            raise ValueError(
                'Invalid data type for box data: float is required.')
        if not self._is_valid_boxes(data):
            raise ValueError('Invalid box data. data must be a numpy array of '
                             'N*[y_min, x_min, y_max, x_max]')
        self.data = {'boxes': data}

    def num_boxes(self) -> int:
        """Return number of boxes held in collections."""
        return self.data['boxes'].shape[0]

    def get_extra_fields(self) -> List[str]:
        """Return all non-box fields."""
        return [k for k in self.data.keys() if k != 'boxes']

    def has_field(self, field) -> bool:
        return field in self.data

    def add_field(self, name: str, data: np.ndarray) -> None:
        """Add data to a specified field.

        Args:
            name (str): Field name.
            data (np.ndarray): Field data: box coords as a [N, 4] numpy array.

        Raises:
            ValueError: If name already exists.
            ValueError: If the dimension of the field data does not matche the
                number of boxes.
        """
        if self.has_field(name):
            raise ValueError('Field ' + name + 'already exists')

        if len(data.shape) < 1 or len(data) != self.num_boxes():
            raise ValueError('Invalid dimensions for name data')

        self.data[name] = data

    def get(self):
        """Shorthand for get_field('boxes')."""
        return self.get_field('boxes')

    def get_field(self, name: str) -> np.ndarray:
        """Get data for field.

        Args:
            name (str): Field name.

        Returns:
            np.ndarray: The data associated with the field.

        Raises:
            ValueError: if invalid field.
        """
        try:
            return self.data[name]
        except KeyError:
            raise ValueError(f'field {name} does not exist')

    def get_coordinates(
            self) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Get corner coordinates of boxes.

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]: a 4-tuple
                of 1-d numpy arrays [y_min, x_min, y_max, x_max].
        """
        boxes = self.get_field('boxes')
        return tuple(boxes.T)

    def _is_valid_boxes(self, data: np.ndarray) -> bool:
        """Check whether data fullfills the format of N*[ymin, xmin, ymax, xmin].

        Args:
            data (np.ndarray): Box coords as a [N, 4] numpy array.

        Returns:
            bool: Whether ymin <= ymax and xmin <= xmax.
        """
        ymin, xmin, ymax, xmax = data.T
        return (ymin <= ymax).all() and (xmin <= xmax).all()



from rastervision.core.data.raster_source import (RasterizedSource)
from rastervision.core.data.vector_source import (VectorSourceConfig)
from rastervision.core.data.vector_transformer import (
    ClassInferenceTransformerConfig, BufferTransformerConfig)
from rastervision.pipeline.config import (register_config, Config, Field,
                                          validator)


@register_config('rasterizer')
class RasterizerConfig(Config):
    """Configure rasterization params for a :class:`.RasterizedSource`."""

    background_class_id: int = Field(
        ...,
        description='The class_id to use for any background pixels, i.e. '
        'pixels not covered by a polygon.')
    all_touched: bool = Field(
        False,
        description='If True, all pixels touched by geometries will be burned '
        'in. If false, only pixels whose center is within the polygon or that '
        'are selected by Bresenham\'s line algorithm will be burned in. '
        '(See rasterio.features.rasterize for more details).')


@register_config('rasterized_source')
class RasterizedSourceConfig(Config):
    """Configure a :class:`.RasterizedSource`."""

    vector_source: VectorSourceConfig
    rasterizer_config: RasterizerConfig

    @validator('vector_source')
    def ensure_required_transformers(
            cls, v: VectorSourceConfig) -> VectorSourceConfig:
        """Add class-inference and buffer transformers if absent."""
        tfs = v.transformers

        # add class inference transformer
        has_inf_tf = any(
            isinstance(tf, ClassInferenceTransformerConfig) for tf in tfs)
        if not has_inf_tf:
            tfs += [ClassInferenceTransformerConfig(default_class_id=None)]

        # add buffer transformers
        has_buf_tf = any(isinstance(tf, BufferTransformerConfig) for tf in tfs)
        if not has_buf_tf:
            tfs += [
                BufferTransformerConfig(geom_type='Point', default_buf=1),
                BufferTransformerConfig(geom_type='LineString', default_buf=1)
            ]

        return v

    def update(self, pipeline=None, scene=None):
        super().update(pipeline, scene)
        self.vector_source.update(pipeline, scene)

    def build(self, class_config, crs_transformer, extent):
        vector_source = self.vector_source.build(class_config, crs_transformer)
        return RasterizedSource(
            vector_source=vector_source,
            background_class_id=self.rasterizer_config.background_class_id,
            extent=extent,
            all_touched=self.rasterizer_config.all_touched)



###############################################################################
# Ported over from https://github.com/azavea/mask-to-polygons.
###############################################################################

from typing import TYPE_CHECKING, Iterator, Optional, Tuple
from itertools import chain

import numpy as np
import cv2
import rasterio as rio
from shapely.geometry import shape

if TYPE_CHECKING:
    from shapely.geometry.base import BaseGeometry

RotatedRectange = Tuple[Tuple[float, float], Tuple[float, float], float]


def mask_to_polygons(mask: np.ndarray, transform: Optional[rio.Affine] = None
                     ) -> Iterator['BaseGeometry']:
    """Polygonize a raster mask. Wrapper around rasterio.features.shapes.

    Args:
        mask (np.ndarray): The mask containing buildings to polygonize.
        transform (Optional[rio.Affine]): Affine transform to use during
            polygonization. Deafults to None (i.e. identity transform).

    Returns:
        Iterator[BaseGeometry]: Generator of shapely polygons.
    """
    if transform is None:
        transform = rio.Affine.identity()
    shapes = rio.features.shapes(mask, mask=(mask == 1), transform=transform)
    polygons = (shape(s) for s, v in shapes if v == 1)
    return polygons


def mask_to_building_polygons(
        mask: np.ndarray,
        transform: Optional[rio.Affine] = None,
        min_area: float = 100,
        width_factor: float = 0.5,
        thickness: float = 0.001) -> Iterator['BaseGeometry']:
    """Try to break up building clusters and then convert to polygons.

    Perofrms the following steps:

    1.  Identify connected components in ``mask``.
    2.  For each connected component, if >= ``min_area``:

        a.  Generate a kernel based on its dimensions and orientation and
            the ``width_factor`` and ``thickness`` params.
        b.  Use the kernel to apply morphological erosion to component-
            mask.
        c.  Identify connected sub-components in component-mask.
        d.  For each connected sub-component, if >= ``min_area``:

            1. Apply morphological dilation using the kernel from above.
            2. Polygonize using ``mask_to_polygons()``.

    Args:
        mask (np.ndarray): The mask containing buildings to polygonize.
        transform (Optional[rio.Affine]): Affine transform to use during
            polygonization. Deafults to None (i.e. identity transform).
        min_area (float): Minimum area (in pixels^2) of anything that can be
            considered to be a building or cluster of buildings. The goal is to
            distinguish between buildings and artifacts. Components with area
            less than this value will be discarded. Defaults to 100.
        width_factor (float): Width of the structural element used to break
            building clusters as a fraction of the width of the cluster.
        thickness (float): Thickness of the structural element that is used to
            break building clusters. Defaults to 0.001.

    Returns:
        Iterator[BaseGeometry]: Generator of shapely polygons.
    """
    n, components = cv2.connectedComponents(mask)

    iterators = []
    for i in range(1, n):
        component = (components == i).astype(np.uint8)
        if component.sum() < min_area:
            continue

        rectangle = get_rectangle(component)
        kernel = get_kernel(rectangle, width_factor, thickness)
        if kernel is None:
            iterators.append(mask_to_polygons(component, transform))
            continue
        eroded = cv2.morphologyEx(
            component, cv2.MORPH_ERODE, kernel, iterations=1)
        m, sub_components = cv2.connectedComponents(eroded)

        for j in range(1, m):
            sub_component = (sub_components == j).astype(np.uint8)
            if sub_component.sum() < min_area:
                continue
            sub_component_dilated = cv2.morphologyEx(
                sub_component, cv2.MORPH_DILATE, kernel, iterations=1)
            iterators.append(
                mask_to_polygons(sub_component_dilated, transform))

    return chain.from_iterable(iterators)


def get_rectangle(buildings: np.ndarray) -> Optional[RotatedRectange]:
    contours, _ = cv2.findContours(buildings, cv2.RETR_EXTERNAL,
                                   cv2.CHAIN_APPROX_SIMPLE)
    if len(contours) > 0:
        rectangle = cv2.minAreaRect(contours[0])
        return rectangle
    else:
        return None


def get_kernel(rectangle: RotatedRectange,
               width_factor: float = 0.5,
               thickness: float = 0.001) -> Optional[np.ndarray]:
    ((cx, cy), (xwidth, ywidth), angle) = rectangle

    width = int(width_factor * min(xwidth, ywidth))

    kernel = np.zeros((width, width), dtype=np.uint8)
    try:
        kernel = cv2.cvtColor(kernel, cv2.COLOR_GRAY2BGR)
    except Exception:
        return None

    diagonal = width * np.sqrt(2)
    pos = (width // 2, width // 2)
    dim = (diagonal, thickness)
    if ywidth < xwidth:
        angle += 90
    element_rect: RotatedRectange = (pos, dim, angle)
    element_contour = cv2.boxPoints(box=element_rect)
    # https://stackoverflow.com/questions/48350693/what-is-numpy-method-int0
    element_contour = np.int0(element_contour)

    cv2.drawContours(
        image=kernel,
        contours=[element_contour],
        contourIdx=0,
        color=(1, 0, 0),
        # -1 means fill interior
        thickness=-1)
    kernel = kernel[:, :, 0]

    return kernel


# Adapted from https://github.com/mapbox/robosat/blob/a8e0e3d676b454b61df03897e43e003867b6ef48/robosat/features/core.py#L65-L77  # noqa
def denoise(mask: np.ndarray, radius: int) -> np.ndarray:
    """Apply morphological opening /w circular kernel to remove hi-freq noise.

    Args:
        mask (np.ndarray): the binary mask to remove noise from.
        radius (int): size in pixels of kernel for morphological op.

    Returns:
        np.ndarray: The mask after applying denoising.
    """
    kernel = cv2.getStructuringElement(
        shape=cv2.MORPH_ELLIPSE, ksize=(radius, radius))
    out = cv2.morphologyEx(src=mask, op=cv2.MORPH_OPEN, kernel=kernel)
    return out



from typing import TYPE_CHECKING, Any, Iterable
from abc import (ABC, abstractmethod)

if TYPE_CHECKING:
    from rastervision.core.data import Scene, Labels


class Evaluator(ABC):
    """Evaluates predictions for a set of scenes."""

    @abstractmethod
    def process(self, scenes: Iterable['Scene']) -> None:
        """Evaluate all given scenes and save the evaluations.

        Args:
            scenes (Iterable[Scene]): Scenes to evaluate.
        """
        pass

    @abstractmethod
    def evaluate_scene(self, scene: 'Scene') -> Any:
        """Evaluate predictions from a scene's labels store.

        The predictions are evalated against ground truth labels from the
        scene's label source.

        Args:
            scene (Scene): A scene with a label source and a label store.

        Returns:
            ClassificationEvaluation: The evaluation.
        """
        pass

    @abstractmethod
    def evaluate_predictions(self, ground_truth: 'Labels',
                             predictions: 'Labels') -> Any:
        """Evaluate predictions against ground truth.

        Args:
            ground_truth (Labels): Ground truth labels.
            predictions (Labels): Predictions.

        Returns:
            Any: The evaluation.
        """
        pass



from abc import abstractmethod
from typing import TYPE_CHECKING, Optional

from rastervision.pipeline.config import Config, register_config

if TYPE_CHECKING:
    from rastervision.core.rv_pipeline import RVPipelineConfig
    from rastervision.core.data import ClassConfig, SceneConfig
    from rastervision.core.data.vector_transformer import VectorTransformer


@register_config('vector_transformer')
class VectorTransformerConfig(Config):
    """Configure a :class:`.VectorTransformer`."""

    def update(self,
               pipeline: Optional['RVPipelineConfig'] = None,
               scene: Optional['SceneConfig'] = None) -> None:
        pass

    @abstractmethod
    def build(self, class_config: 'ClassConfig') -> 'VectorTransformer':
        pass



from typing import TYPE_CHECKING, Optional
import numpy as np

from shapely.ops import transform

from rastervision.core.data.crs_transformer import RasterioCRSTransformer
from rastervision.core.data.utils.geojson import (
    pixel_to_map_coords, map_to_pixel_coords, map_geoms)
from rastervision.core.data.vector_transformer import VectorTransformer

if TYPE_CHECKING:
    from rastervision.core.data import CRSTransformer

METERS_PER_DEGREE = 111319.5
RADIANS_PER_DEGREE = np.pi / 180


class ShiftTransformer(VectorTransformer):
    """Shift geometries by some distance specified in meters."""

    def __init__(self,
                 x_shift: float = 0.,
                 y_shift: float = 0.,
                 round_pixels: bool = True):
        """Constructor.

        Args:
        """
        self.x_shift = x_shift
        self.y_shift = y_shift
        self.round_pixels = round_pixels

    def transform(self,
                  geojson: dict,
                  crs_transformer: Optional['CRSTransformer'] = None) -> dict:

        # https://gis.stackexchange.com/questions/2951/algorithm-for-offsetting-a-latitude-longitude-by-some-amount-of-meters  # noqa
        def shift(x, y, z=None):
            lon, lat = x, y
            lat_radians = RADIANS_PER_DEGREE * lat
            dlon = self.x_shift / (METERS_PER_DEGREE * np.cos(lat_radians))
            dlat = self.y_shift / METERS_PER_DEGREE
            return lon + dlon, lat + dlat

        wgs84_transformer = self.make_wgs84_transformer(crs_transformer)

        geojson_pixel = geojson
        geojson_wgs84 = pixel_to_map_coords(geojson_pixel, wgs84_transformer)
        geojson_wgs84_shifted = map_geoms(lambda g, **kw: transform(shift, g),
                                          geojson_wgs84)
        geojson_pixel_shifted = map_to_pixel_coords(geojson_wgs84_shifted,
                                                    wgs84_transformer)

        return geojson_pixel_shifted

    def make_wgs84_transformer(self, crs_transformer: 'CRSTransformer'):
        wgs84_transformer = RasterioCRSTransformer(
            transform=crs_transformer.transform,
            image_crs=crs_transformer.image_crs,
            map_crs='epsg:4326',
            round_pixels=self.round_pixels)
        return wgs84_transformer



import numpy as np


def nodata_below_threshold(chip: np.ndarray,
                           threshold: float,
                           nodata_val: int = 0) -> bool:
    """Check if proportion of nodata pixels is below the threshold.

    Args:
        chip (np.ndarray): Raster as (H, W) or (H, W, C) numpy array.
        threshold (float): Threshold to check the fraction of NODATA pixels
            against.
        nodata_val (int, optional): Value that represents NODATA pixels.
            Defaults to 0.

    Returns:
        bool: Whether the fraction of NODATA pixels is below the given
            threshold.
    """
    if len(chip.shape) == 3:
        # (H, W, C) --> (H, W)
        chip = chip.sum(axis=-1)
    nodata_prop = (chip == nodata_val).mean()
    return nodata_prop < threshold



from typing import TYPE_CHECKING, Callable, List, Optional
import logging

from rastervision.core.rv_pipeline.rv_pipeline import RVPipeline
from rastervision.core.rv_pipeline.utils import nodata_below_threshold
from rastervision.core.rv_pipeline.object_detection_config import (
    ObjectDetectionWindowMethod)
from rastervision.core.box import Box
from rastervision.core.data.label import ObjectDetectionLabels

if TYPE_CHECKING:
    from rastervision.core.backend.backend import Backend
    from rastervision.core.data import (Labels, Scene, RasterSource,
                                        ObjectDetectionLabelSource)
    from rastervision.core.rv_pipeline.object_detection_config import (
        ObjectDetectionChipOptions)

log = logging.getLogger(__name__)


def _make_chip_pos_windows(image_extent: Box,
                           label_source: 'ObjectDetectionLabelSource',
                           chip_size: int) -> List[Box]:
    chip_size = chip_size
    pos_windows = []
    boxes = label_source.get_labels().get_boxes()
    done_boxes = set()

    # Get a random window around each box. If a box was previously included
    # in a window, then it is skipped.
    for box in boxes:
        if box in done_boxes:
            continue
        # If this  object is bigger than the chip, don't use this box.
        if chip_size < box.width or chip_size < box.height:
            log.warning(f'Label is larger than chip size: {box} '
                        'Skipping this label.')
            continue

        window = box.make_random_square_container(chip_size)
        pos_windows.append(window)

        # Get boxes that lie completely within window
        window_boxes = label_source.get_labels(window=window)
        window_boxes = ObjectDetectionLabels.get_overlapping(
            window_boxes, window, ioa_thresh=1.0)
        window_boxes = window_boxes.get_boxes()
        done_boxes.update(window_boxes)

    return pos_windows


def _make_label_pos_windows(image_extent: Box,
                            label_source: 'ObjectDetectionLabelSource',
                            label_buffer: int) -> List[Box]:
    pos_windows = []
    boxes = label_source.get_labels().get_boxes()
    for box in boxes:
        window = box.buffer(label_buffer, image_extent)
        pos_windows.append(window)

    return pos_windows


def make_pos_windows(
        image_extent: Box, label_source: 'ObjectDetectionLabelSource',
        chip_size: int, window_method: ObjectDetectionWindowMethod,
        label_buffer: Optional[int]) -> List[Box]:
    if window_method == ObjectDetectionWindowMethod.chip:
        return _make_chip_pos_windows(image_extent, label_source, chip_size)
    elif window_method == ObjectDetectionWindowMethod.label:
        if label_buffer is None:
            raise ValueError(
                'label_buffer must be specified if '
                'window_method=ObjectDetectionWindowMethod.label.')
        return _make_label_pos_windows(image_extent, label_source,
                                       label_buffer)
    elif window_method == ObjectDetectionWindowMethod.image:
        return [image_extent.copy()]
    else:
        raise NotImplementedError(f'Window method: {window_method}.')


def make_neg_windows(raster_source: 'RasterSource',
                     label_source: 'ObjectDetectionLabelSource',
                     chip_size: int,
                     nb_windows: int,
                     max_attempts: int,
                     filter_windows: Callable,
                     chip_nodata_threshold: float = 1.) -> List[Box]:
    extent = raster_source.extent
    neg_windows = []
    for _ in range(max_attempts):
        for _ in range(max_attempts):
            window = extent.make_random_square(chip_size)
            if any(filter_windows([window])):
                break
        chip = raster_source.get_chip(window)
        labels = ObjectDetectionLabels.get_overlapping(
            label_source.get_labels(), window, ioa_thresh=0.2)

        # If no labels and not too many nodata pixels, append the chip
        nodata_below_thresh = nodata_below_threshold(
            chip, chip_nodata_threshold, nodata_val=0)
        if len(labels) == 0 and nodata_below_thresh:
            neg_windows.append(window)

        if len(neg_windows) == nb_windows:
            break

    return list(neg_windows)


def get_train_windows(scene: 'Scene',
                      chip_opts: 'ObjectDetectionChipOptions',
                      chip_size: int,
                      chip_nodata_threshold: float = 1.) -> List[Box]:
    raster_source = scene.raster_source
    label_source = scene.label_source

    def filter_windows(windows):
        if scene.aoi_polygons:
            windows = Box.filter_by_aoi(windows, scene.aoi_polygons)
        return windows

    window_method = chip_opts.window_method
    if window_method == ObjectDetectionWindowMethod.sliding:
        stride = chip_size
        windows = raster_source.extent.get_windows(chip_size, stride)
        return list(filter_windows(windows))

    # Make positive windows which contain labels.
    pos_windows = make_pos_windows(raster_source.extent, label_source,
                                   chip_size, chip_opts.window_method,
                                   chip_opts.label_buffer)
    pos_windows = filter_windows(pos_windows)
    nb_pos_windows = len(pos_windows)

    # Make negative windows which do not contain labels.
    # Generate randow windows and save the ones that don't contain
    # any labels. It may take many attempts to generate a single
    # negative window, and could get into an infinite loop in some cases,
    # so we cap the number of attempts.
    if nb_pos_windows:
        nb_neg_windows = round(chip_opts.neg_ratio * nb_pos_windows)
    else:
        nb_neg_windows = 100  # just make some
    max_attempts = 100 * nb_neg_windows
    neg_windows = make_neg_windows(
        raster_source,
        label_source,
        chip_size,
        nb_neg_windows,
        max_attempts,
        filter_windows,
        chip_nodata_threshold=chip_nodata_threshold)

    return pos_windows + neg_windows


class ObjectDetection(RVPipeline):
    def get_train_windows(self, scene: 'Scene') -> List[Box]:
        return get_train_windows(
            scene,
            self.config.chip_options,
            self.config.train_chip_sz,
            chip_nodata_threshold=self.config.chip_nodata_threshold)

    def get_train_labels(self, window: Box,
                         scene: 'Scene') -> ObjectDetectionLabels:
        window_labels = scene.label_source.get_labels(window=window)
        return ObjectDetectionLabels.get_overlapping(
            window_labels,
            window,
            ioa_thresh=self.config.chip_options.ioa_thresh,
            clip=True)

    def predict_scene(self, scene: 'Scene', backend: 'Backend') -> 'Labels':
        # Use strided windowing to ensure that each object is fully visible (ie. not
        # cut off) within some window. This means prediction takes 4x longer for object
        # detection :(
        chip_sz = self.config.predict_chip_sz
        stride = chip_sz // 2
        return backend.predict_scene(scene, chip_sz=chip_sz, stride=stride)

    def post_process_predictions(self, labels: ObjectDetectionLabels,
                                 scene: 'Scene') -> ObjectDetectionLabels:
        return ObjectDetectionLabels.prune_duplicates(
            labels,
            score_thresh=self.config.predict_options.score_thresh,
            merge_thresh=self.config.predict_options.merge_thresh)



from typing import TYPE_CHECKING

from rastervision.core.data.label import ObjectDetectionLabels
from rastervision.core.data.label_store import LabelStore
from rastervision.core.data.label_store.utils import boxes_to_geojson
from rastervision.core.data.vector_source import GeoJSONVectorSourceConfig
from rastervision.pipeline.file_system import json_to_file

if TYPE_CHECKING:
    from rastervision.core.data import ClassConfig, CRSTransformer


class ObjectDetectionGeoJSONStore(LabelStore):
    """Storage for object detection predictions."""

    def __init__(self, uri: str, class_config: 'ClassConfig',
                 crs_transformer: 'CRSTransformer'):
        """Constructor.

        Args:
            uri: uri of GeoJSON file containing labels
            class_config: ClassConfig used to infer class_ids from class_name
                (or label) field
            crs_transformer: CRSTransformer to convert from map coords in label
                in GeoJSON file to pixel coords.
        """
        self.uri = uri
        self.crs_transformer = crs_transformer
        self.class_config = class_config

    def save(self, labels: ObjectDetectionLabels) -> None:
        """Save labels to URI."""
        boxes = labels.get_boxes()
        class_ids = [int(id) for id in labels.get_class_ids()]
        scores = labels.get_scores()
        geojson = boxes_to_geojson(
            boxes,
            class_ids,
            self.crs_transformer,
            self.class_config,
            scores=scores)
        json_to_file(geojson, self.uri)

    def get_labels(self) -> ObjectDetectionLabels:
        vector_source = GeoJSONVectorSourceConfig(uri=self.uri).build(
            self.class_config, self.crs_transformer)
        return ObjectDetectionLabels.from_geojson(vector_source.get_geojson())

    def empty_labels(self) -> ObjectDetectionLabels:
        return ObjectDetectionLabels.make_empty()



from typing import TYPE_CHECKING, Any, Optional, Tuple

if TYPE_CHECKING:
    from rastervision.core.box import Box
    from rastervision.core.data import (RasterSource, LabelSource, LabelStore)


class Scene:
    """The raster data and labels associated with an area of interest."""

    def __init__(self,
                 id: str,
                 raster_source: 'RasterSource',
                 label_source: Optional['LabelSource'] = None,
                 label_store: Optional['LabelStore'] = None,
                 aoi_polygons: Optional[list] = None):
        """Construct a new Scene.

        Args:
            id: ID for this scene
            raster_source: RasterSource for this scene
            ground_truth_label_store: optional LabelSource
            label_store: optional LabelStore
            aoi: Optional list of AOI polygons in pixel coordinates
        """
        self.id = id
        self.raster_source = raster_source
        self.label_source = label_source
        self.label_store = label_store
        if aoi_polygons is None:
            self.aoi_polygons = []
        else:
            self.aoi_polygons = aoi_polygons

    @property
    def extent(self) -> 'Box':
        """Extent of the associated :class:`.RasterSource`."""
        return self.raster_source.extent

    def __getitem__(self, key: Any) -> Tuple[Any, Any]:
        x = self.raster_source[key]
        if self.label_source is not None:
            y = self.label_source[key]
        else:
            y = None
        return x, y



# flake8: noqa

from rastervision.core.data.vector_source.vector_source import *
from rastervision.core.data.vector_source.vector_source_config import *
from rastervision.core.data.vector_source.geojson_vector_source import *
from rastervision.core.data.vector_source.geojson_vector_source_config import *

__all__ = [
    VectorSource.__name__,
    VectorSourceConfig.__name__,
    GeoJSONVectorSource.__name__,
    GeoJSONVectorSourceConfig.__name__,
]



from abc import abstractmethod
from typing import TYPE_CHECKING, List, Optional

from rastervision.pipeline.config import Config, register_config, Field
from rastervision.core.data.vector_transformer import VectorTransformerConfig

if TYPE_CHECKING:
    from rastervision.core.rv_pipeline import RVPipelineConfig
    from rastervision.core.data import (ClassConfig, CRSTransformer,
                                        SceneConfig, VectorSource)


def vector_source_config_upgrader(cfg_dict: dict, version: int) -> dict:
    if version == 4:
        from rastervision.core.data.vector_transformer import (
            ClassInferenceTransformerConfig, BufferTransformerConfig)

        class_inf_tf = ClassInferenceTransformerConfig(
            default_class_id=cfg_dict.get('default_class_id'),
            class_id_to_filter=cfg_dict.get('class_id_to_filter')).dict()

        line_bufs = {} if cfg_dict.get('line_bufs') is None else cfg_dict.get(
            'line_bufs')
        point_bufs = {} if cfg_dict.get(
            'point_bufs') is None else cfg_dict.get('point_bufs')
        buf_tfs = [
            BufferTransformerConfig(
                geom_type='LineString', class_bufs=line_bufs).dict(),
            BufferTransformerConfig(geom_type='Point',
                                    class_bufs=point_bufs).dict()
        ]
        # added in version 5
        cfg_dict['transformers'] = [class_inf_tf, *buf_tfs]
        try:
            # removed in version 5
            del cfg_dict['default_class_id']
            del cfg_dict['class_id_to_filter']
            del cfg_dict['line_bufs']
            del cfg_dict['point_bufs']
        except KeyError:
            pass
    return cfg_dict


@register_config('vector_source', upgrader=vector_source_config_upgrader)
class VectorSourceConfig(Config):
    """Configure a :class:`.VectorSource`."""

    transformers: List[VectorTransformerConfig] = Field(
        [], description='List of VectorTransformers.')

    @abstractmethod
    def build(self, class_config: 'ClassConfig',
              crs_transformer: 'CRSTransformer') -> 'VectorSource':
        pass

    def update(self,
               pipeline: Optional['RVPipelineConfig'] = None,
               scene: Optional['SceneConfig'] = None) -> None:
        pass



# flake8: noqa

from rastervision.core.analyzer.analyzer import *
from rastervision.core.analyzer.analyzer_config import *
from rastervision.core.analyzer.stats_analyzer import *
from rastervision.core.analyzer.stats_analyzer_config import *

__all__ = [
    Analyzer.__name__,
    AnalyzerConfig.__name__,
    StatsAnalyzer.__name__,
    StatsAnalyzerConfig.__name__,
]



from typing import TYPE_CHECKING, List, Optional, Union
from uuid import uuid4

from rastervision.core.data.utils import listify_uris, get_polygons_from_uris

if TYPE_CHECKING:
    from rastervision.core.data import ClassConfig, Scene


def make_ss_scene(image_uri: Union[str, List[str]],
                  label_raster_uri: Optional[Union[str, List[str]]] = None,
                  class_config: Optional['ClassConfig'] = None,
                  label_vector_uri: Optional[str] = None,
                  aoi_uri: Union[str, List[str]] = [],
                  label_vector_default_class_id: Optional[int] = None,
                  image_raster_source_kw: dict = {},
                  label_raster_source_kw: dict = {},
                  label_vector_source_kw: dict = {},
                  scene_id: Optional[str] = None) -> 'Scene':
    """Create a semantic segmentation scene from image and label URIs.

    This is a convenience method. For more fine-grained control, it is
    recommended to use the default constructor.

    Args:
        image_uri (Union[str, List[str]]): URI or list of URIs of GeoTIFFs to
            use as the source of image data.
        label_raster_uri (Optional[Union[str, List[str]]], optional): URI or
            list of URIs of GeoTIFFs to use as the source of segmentation label
            data. If the labels are in the form of GeoJSONs, use
            label_vector_uri instead. Defaults to None.
        label_vector_uri (Optional[str], optional):  URI of GeoJSON file to use
            as the source of segmentation label data. If the labels are in the
            form of GeoTIFFs, use label_raster_uri instead. Defaults to None.
        class_config (Optional[ClassConfig]): The ClassConfig. Must be
            non-None if creating a scene without a LabelSource.
            Defaults to None.
        aoi_uri (Union[str, List[str]], optional): URI or list of URIs of
            GeoJSONs that specify the area-of-interest. If provided, the
            dataset will only access data from this area. Defaults to [].
        label_vector_default_class_id (Optional[int], optional): If using
            label_vector_uri and all polygons in that file belong to the same
            class and they do not contain a `class_id` property, then use this
            argument to map all of the polgons to the appropriate class ID.
            See docs for ClassInferenceTransformer for more details.
            Defaults to None.
        image_raster_source_kw (dict, optional): Additional arguments to pass
            to the RasterioSource used for image data. See docs for
            RasterioSource for more details. Defaults to {}.
        label_raster_source_kw (dict, optional): Additional arguments to pass
            to the RasterioSource used for label data, if label_raster_uri is
            used. See docs for RasterioSource for more details. Defaults to {}.
        label_vector_source_kw (dict, optional): Additional arguments to pass
            to the GeoJSONVectorSource used for label data, if label_vector_uri
            is used. See docs for GeoJSONVectorSource for more details.
            Defaults to {}.
        scene_id (Optional[str]): Optional scene ID. If None, will be randomly
            generated. Defaults to None.

    Raises:
        ValueError: If both label_raster_uri and label_vector_uri are
            specified.

    Returns:
        Scene: A semantic segmentation scene.
    """
    # use local imports to avoid circular import problems
    from rastervision.core.data import (
        GeoJSONVectorSource, RasterioSource, RasterizedSource, Scene,
        SemanticSegmentationLabelSource, ClassInferenceTransformer)

    if label_raster_uri is not None and label_vector_uri is not None:
        raise ValueError('Specify either label_raster_uri or '
                         'label_vector_uri or neither, but not both.')

    if label_raster_uri is not None or label_vector_uri is not None:
        if class_config is None:
            raise ValueError('class_config is required if using labels.')
        class_config.ensure_null_class()

    image_uri = listify_uris(image_uri)
    raster_source = RasterioSource(uris=image_uri, **image_raster_source_kw)

    crs_transformer = raster_source.crs_transformer
    extent = raster_source.extent

    label_raster_source = None
    if label_raster_uri is not None:
        label_raster_uri = listify_uris(label_raster_uri)
        label_raster_source = RasterioSource(
            uris=label_raster_uri, **label_raster_source_kw)
    elif label_vector_uri is not None:
        if label_vector_default_class_id is not None:
            # add a ClassInferenceTransformer to the VectorSource
            class_inf_tf = ClassInferenceTransformer(
                default_class_id=label_vector_default_class_id)
            vector_tfs = label_vector_source_kw.get('vector_transformers', [])
            label_vector_source_kw['vector_transformers'] = (
                [class_inf_tf] + vector_tfs)
        vector_source = GeoJSONVectorSource(
            uri=label_vector_uri,
            ignore_crs_field=True,
            crs_transformer=crs_transformer,
            **label_vector_source_kw)
        label_raster_source = RasterizedSource(
            vector_source=vector_source,
            background_class_id=label_raster_source_kw.pop(
                'background_class_id', class_config.null_class_id),
            extent=extent,
            **label_raster_source_kw)

    label_source = None
    if label_raster_source is not None:
        label_source = SemanticSegmentationLabelSource(
            raster_source=label_raster_source, class_config=class_config)

    aoi_polygons = get_polygons_from_uris(aoi_uri, crs_transformer)
    scene = Scene(
        id=uuid4() if scene_id is None else scene_id,
        raster_source=raster_source,
        label_source=label_source,
        aoi_polygons=aoi_polygons)

    return scene


def make_cc_scene(image_uri: Union[str, List[str]],
                  label_vector_uri: Optional[str] = None,
                  class_config: Optional['ClassConfig'] = None,
                  aoi_uri: Union[str, List[str]] = [],
                  label_vector_default_class_id: Optional[int] = None,
                  image_raster_source_kw: dict = {},
                  label_vector_source_kw: dict = {},
                  label_source_kw: dict = {},
                  scene_id: Optional[str] = None) -> 'Scene':
    """Create a chip classification scene from image and label URIs.

    This is a convenience method. For more fine-grained control, it is
    recommended to use the default constructor.

    Args:
        image_uri (Union[str, List[str]]): URI or list of URIs of GeoTIFFs to
            use as the source of image data.
        label_vector_uri (Optional[str], optional):  URI of GeoJSON file to use
            as the source of segmentation label data. Defaults to None.
        class_config (Optional[ClassConfig]): The ClassConfig. Must be
            non-None if creating a scene without a LabelSource.
            Defaults to None.
        aoi_uri (Union[str, List[str]], optional): URI or list of URIs of
            GeoJSONs that specify the area-of-interest. If provided, the
            dataset will only access data from this area. Defaults to [].
        label_vector_default_class_id (Optional[int], optional): If using
            label_vector_uri and all polygons in that file belong to the same
            class and they do not contain a `class_id` property, then use this
            argument to map all of the polgons to the appropriate class ID.
            See docs for ClassInferenceTransformer for more details.
            Defaults to None.
        image_raster_source_kw (dict, optional): Additional arguments to pass
            to the RasterioSource used for image data. See docs for
            RasterioSource for more details. Defaults to {}.
        label_vector_source_kw (dict, optional): Additional arguments to pass
            to the GeoJSONVectorSourceConfig used for label data, if
            label_vector_uri is set. See docs for GeoJSONVectorSourceConfig
            for more details. Defaults to {}.
        label_source_kw (dict, optional): Additional arguments to pass
            to the ChipClassificationLabelSourceConfig used for label data, if
            label_vector_uri is set. See docs for
            ChipClassificationLabelSourceConfig for more details.
            Defaults to {}.
        scene_id (Optional[str]): Optional scene ID. If None, will be randomly
            generated. Defaults to None.

    Returns:
        Scene: A chip classification scene.
    """
    # use local imports to avoid circular import problems
    from rastervision.core.data import (
        RasterioSource, Scene, ClassInferenceTransformerConfig,
        ChipClassificationLabelSourceConfig, GeoJSONVectorSourceConfig)

    image_uri = listify_uris(image_uri)
    raster_source = RasterioSource(image_uri, **image_raster_source_kw)

    crs_transformer = raster_source.crs_transformer
    extent = raster_source.extent

    label_source = None
    if label_vector_uri is not None:
        if class_config is None:
            raise ValueError('class_config is required if using labels.')
        if label_vector_default_class_id is not None:
            # add a ClassInferenceTransformer to the VectorSource
            class_inf_tf = ClassInferenceTransformerConfig(
                default_class_id=label_vector_default_class_id)
            vector_tfs = label_vector_source_kw.get('transformers', [])
            label_vector_source_kw['transformers'] = (
                [class_inf_tf] + vector_tfs)
        geojson_cfg = GeoJSONVectorSourceConfig(
            uri=label_vector_uri,
            ignore_crs_field=True,
            **label_vector_source_kw)
        # use config to ensure required transformers are auto added
        label_source_cfg = ChipClassificationLabelSourceConfig(
            vector_source=geojson_cfg, **label_source_kw)
        label_source = label_source_cfg.build(
            class_config, crs_transformer, extent=extent)

    aoi_polygons = get_polygons_from_uris(aoi_uri, crs_transformer)
    scene = Scene(
        id=uuid4() if scene_id is None else scene_id,
        raster_source=raster_source,
        label_source=label_source,
        aoi_polygons=aoi_polygons)

    return scene


def make_od_scene(image_uri: Union[str, List[str]],
                  label_vector_uri: Optional[str] = None,
                  class_config: Optional['ClassConfig'] = None,
                  aoi_uri: Union[str, List[str]] = [],
                  label_vector_default_class_id: Optional[int] = None,
                  image_raster_source_kw: dict = {},
                  label_vector_source_kw: dict = {},
                  label_source_kw: dict = {},
                  scene_id: Optional[str] = None) -> 'Scene':
    """Create an object detection scene from image and label URIs.

    This is a convenience method. For more fine-grained control, it is
    recommended to use the default constructor.

    Args:
        image_uri (Union[str, List[str]]): URI or list of URIs of GeoTIFFs to
            use as the source of image data.
        label_vector_uri (Optional[str], optional):  URI of GeoJSON file to use
            as the source of segmentation label data. Defaults to None.
        class_config (Optional[ClassConfig]): The ClassConfig. Must be
            non-None if creating a scene without a LabelSource.
            Defaults to None.
        aoi_uri (Union[str, List[str]], optional): URI or list of URIs of
            GeoJSONs that specify the area-of-interest. If provided, the
            dataset will only access data from this area. Defaults to [].
        label_vector_default_class_id (Optional[int], optional): If using
            label_vector_uri and all polygons in that file belong to the same
            class and they do not contain a `class_id` property, then use this
            argument to map all of the polgons to the appropriate class ID.
            See docs for ClassInferenceTransformer for more details.
            Defaults to None.
        image_raster_source_kw (dict, optional): Additional arguments to pass
            to the RasterioSource used for image data. See docs for
            RasterioSource for more details. Defaults to {}.
        label_vector_source_kw (dict, optional): Additional arguments to pass
            to the GeoJSONVectorSourceConfig used for label data, if
            label_vector_uri is set. See docs for GeoJSONVectorSourceConfig
            for more details. Defaults to {}.
        label_source_kw (dict, optional): Additional arguments to pass
            to the ObjectDetectionLabelSourceConfig used for label data, if
            label_vector_uri is set. See docs for
            ObjectDetectionLabelSourceConfig for more details.
            Defaults to {}.
        scene_id (Optional[str]): Optional scene ID. If None, will be randomly
            generated. Defaults to None.

    Returns:
        Scene: An object detection scene.
    """
    # use local imports to avoid circular import problems
    from rastervision.core.data import (
        RasterioSource, Scene, ClassInferenceTransformerConfig,
        GeoJSONVectorSourceConfig, ObjectDetectionLabelSourceConfig)

    image_uri = listify_uris(image_uri)
    raster_source = RasterioSource(image_uri, **image_raster_source_kw)

    crs_transformer = raster_source.crs_transformer
    extent = raster_source.extent

    label_source = None
    if label_vector_uri is not None:
        if class_config is None:
            raise ValueError('class_config is required if using labels.')
        if label_vector_default_class_id is not None:
            # add a ClassInferenceTransformer to the VectorSource
            class_inf_tf = ClassInferenceTransformerConfig(
                default_class_id=label_vector_default_class_id)
            vector_tfs = label_vector_source_kw.get('transformers', [])
            label_vector_source_kw['transformers'] = (
                [class_inf_tf] + vector_tfs)
        geojson_cfg = GeoJSONVectorSourceConfig(
            uri=label_vector_uri,
            ignore_crs_field=True,
            **label_vector_source_kw)
        # use config to ensure required transformers are auto added
        label_source_cfg = ObjectDetectionLabelSourceConfig(
            vector_source=geojson_cfg, **label_source_kw)
        label_source = label_source_cfg.build(
            class_config, crs_transformer, extent=extent)

    aoi_polygons = get_polygons_from_uris(aoi_uri, crs_transformer)
    scene = Scene(
        id=uuid4() if scene_id is None else scene_id,
        raster_source=raster_source,
        label_source=label_source,
        aoi_polygons=aoi_polygons)

    return scene



from typing import TYPE_CHECKING, Iterable, Optional, Tuple

from rastervision.pipeline.config import register_config
from rastervision.core.evaluation.classification_evaluator_config import (
    ClassificationEvaluatorConfig)
from rastervision.core.evaluation.semantic_segmentation_evaluator import (
    SemanticSegmentationEvaluator)

if TYPE_CHECKING:
    from rastervision.core.data import ClassConfig


def ss_evaluator_config_upgrader(cfg_dict: dict, version: int) -> dict:
    if version < 3:
        try:
            # removed in version 3
            del cfg_dict['vector_output_uri']
        except KeyError:
            pass
    return cfg_dict


@register_config(
    'semantic_segmentation_evaluator', upgrader=ss_evaluator_config_upgrader)
class SemanticSegmentationEvaluatorConfig(ClassificationEvaluatorConfig):
    """Configure a :class:`.SemanticSegmentationEvaluator`."""

    def build(self,
              class_config: 'ClassConfig',
              scene_group: Optional[Tuple[str, Iterable[str]]] = None
              ) -> SemanticSegmentationEvaluator:
        if scene_group is None:
            output_uri = self.get_output_uri()
        else:
            group_name, _ = scene_group
            output_uri = self.get_output_uri(group_name)

        evaluator = SemanticSegmentationEvaluator(class_config, output_uri)
        return evaluator



from typing import TYPE_CHECKING, Dict, Optional

from rastervision.pipeline.config import register_config, Field
from rastervision.core.data.vector_transformer import (
    VectorTransformerConfig, ClassInferenceTransformer)

if TYPE_CHECKING:
    from rastervision.core.data import ClassConfig


@register_config('class_inference_transformer')
class ClassInferenceTransformerConfig(VectorTransformerConfig):
    """Configure a :class:`.ClassInferenceTransformer`."""

    default_class_id: Optional[int] = Field(
        None,
        description='The default class_id to use if class cannot be inferred '
        'using other mechanisms. If a feature has an inferred class_id of '
        'None, then it will be deleted.')
    class_id_to_filter: Optional[Dict[int, list]] = Field(
        None,
        description='Map from class_id to JSON filter used to infer missing '
        'class_ids. Each key should be a class id, and its value should be a '
        'boolean expression which is run against the property field for each '
        'feature. This allows matching different features to different '
        'class IDs based on its properties. The expression schema is that '
        'described by '
        'https://docs.mapbox.com/mapbox-gl-js/style-spec/other/#other-filter')

    def build(self, class_config: Optional['ClassConfig'] = None
              ) -> ClassInferenceTransformer:
        return ClassInferenceTransformer(
            self.default_class_id,
            class_config=class_config,
            class_id_to_filter=self.class_id_to_filter)



from typing import TYPE_CHECKING, Optional, List
from os.path import join

from rastervision.core.data.label_store import (LabelStoreConfig,
                                                SemanticSegmentationLabelStore)
from rastervision.pipeline.config import register_config, Config, Field

if TYPE_CHECKING:
    from rastervision.core.data import SceneConfig  # noqa
    from rastervision.core.rv_pipeline import RVPipelineConfig  # noqa


@register_config('vector_output')
class VectorOutputConfig(Config):
    """Config for vectorized semantic segmentation predictions."""
    uri: Optional[str] = Field(
        None,
        description='URI of vector output. If None, and this Config is part '
        'of a SceneConfig and RVPipeline, this field will be auto-generated.')
    class_id: int = Field(
        ...,
        description='The prediction class that is to turned into vectors.')
    denoise: int = Field(
        8,
        description='Radius of the structural element used to remove '
        'high-frequency signals from the image. Smaller values will reduce '
        'less noise and make vectorization slower (especially for large '
        'images). Larger values will remove more noise and make vectorization '
        'faster but might also remove legitimate detections.')

    def update(self,
               pipeline: Optional['RVPipelineConfig'] = None,
               scene: Optional['SceneConfig'] = None,
               uri_prefix: Optional[str] = None):
        if self.uri is None:
            mode = self.get_mode()
            class_id = self.class_id
            filename = f'{mode}-{class_id}.json'
            if uri_prefix is not None:
                self.uri = join(uri_prefix, 'vector_output', filename)
            elif pipeline and scene:
                self.uri = join(pipeline.predict_uri, scene.id,
                                'vector_output', filename)

    def get_mode(self) -> str:
        raise NotImplementedError()


@register_config('polygon_vector_output')
class PolygonVectorOutputConfig(VectorOutputConfig):
    """Config for vectorized semantic segmentation predictions."""

    def get_mode(self) -> str:
        return 'polygons'


def building_vo_config_upgrader(cfg_dict: dict, version: int) -> dict:
    if version == 6:
        try:
            # removed in version 7
            del cfg_dict['min_aspect_ratio']
        except KeyError:
            pass
    return cfg_dict


@register_config(
    'building_vector_output', upgrader=building_vo_config_upgrader)
class BuildingVectorOutputConfig(VectorOutputConfig):
    """Config for vectorized semantic segmentation predictions.

    Intended to break up clusters of buildings.
    """
    min_area: float = Field(
        0.0,
        description='Minimum area (in pixels^2) of anything that can be '
        'considered to be a building or a cluster of buildings. The goal is '
        'to distinguish between buildings and artifacts.')
    element_width_factor: float = Field(
        0.5,
        description='Width of the structural element used to break building '
        'clusters as a fraction of the width of the cluster.')
    element_thickness: float = Field(
        0.001,
        description='Thickness of the structural element that is used to '
        'break building clusters.')

    def get_mode(self) -> str:
        return 'buildings'


@register_config('semantic_segmentation_label_store')
class SemanticSegmentationLabelStoreConfig(LabelStoreConfig):
    """Configure a :class:`.SemanticSegmentationLabelStore`.

    Stores class raster as GeoTIFF, and can optionally vectorizes predictions and stores
    them in GeoJSON files.
    """
    uri: Optional[str] = Field(
        None,
        description=(
            'URI of file with predictions. If None, and this Config is part of '
            'a SceneConfig inside an RVPipelineConfig, this fiend will be '
            'auto-generated.'))
    vector_output: List[VectorOutputConfig] = []
    rgb: bool = Field(
        False,
        description=
        ('If True, save prediction class_ids in RGB format using the colors in '
         'class_config.'))
    smooth_output: bool = Field(
        False,
        description='If True, expects labels to be continuous values '
        'representing class scores and stores both scores and discrete '
        'labels.')
    smooth_as_uint8: bool = Field(
        False,
        description='If True, stores smooth scores as uint8, resulting in '
        'loss of precision, but reduced file size. Only used if '
        'smooth_output=True.')
    rasterio_block_size: int = Field(
        256,
        description='blockxsize and blockysize params in rasterio.open() will '
        'be set to this.')

    def build(self, class_config, crs_transformer, extent, tmp_dir):
        class_config.ensure_null_class()

        label_store = SemanticSegmentationLabelStore(
            uri=self.uri,
            extent=extent,
            crs_transformer=crs_transformer,
            class_config=class_config,
            tmp_dir=tmp_dir,
            vector_outputs=self.vector_output,
            save_as_rgb=self.rgb,
            smooth_output=self.smooth_output,
            smooth_as_uint8=self.smooth_as_uint8,
            rasterio_block_size=self.rasterio_block_size)

        return label_store

    def update(self,
               pipeline: Optional['RVPipelineConfig'] = None,
               scene: Optional['SceneConfig'] = None):
        if pipeline is not None and scene is not None:
            if self.uri is None:
                self.uri = join(pipeline.predict_uri, f'{scene.id}')

        for vo in self.vector_output:
            vo.update(pipeline, scene)



from pydantic import conint, conlist

from rastervision.pipeline.config import (Field, register_config, validator)
from rastervision.core.data.raster_source import (RasterSourceConfig,
                                                  MultiRasterSource)


def multi_rs_config_upgrader(cfg_dict: dict, version: int) -> dict:
    if version == 1:
        # field renamed in version 2
        cfg_dict['primary_source_idx'] = cfg_dict.get('crs_source', 0)
        try:
            del cfg_dict['crs_source']
        except KeyError:
            pass
    return cfg_dict


@register_config('multi_raster_source', upgrader=multi_rs_config_upgrader)
class MultiRasterSourceConfig(RasterSourceConfig):
    """Configure a :class:`.MultiRasterSource`."""

    raster_sources: conlist(
        RasterSourceConfig, min_items=1) = Field(
            ..., description='List of RasterSourceConfig to combine.')
    primary_source_idx: conint(ge=0) = Field(
        0,
        description=
        'Index of the raster source whose CRS, dtype, and other attributes '
        'will override those of the other raster sources. Defaults to 0.')
    force_same_dtype: bool = Field(
        False,
        description='Force all subchips to be of the same dtype as the '
        'primary_source_idx-th subchip.')

    @validator('primary_source_idx')
    def validate_primary_source_idx(cls, v: int, values: dict):
        raster_sources = values.get('raster_sources', [])
        if not (0 <= v < len(raster_sources)):
            raise IndexError('primary_source_idx must be in range '
                             '[0, len(raster_sources)].')
        return v

    def build(self, tmp_dir: str,
              use_transformers: bool = True) -> MultiRasterSource:
        if use_transformers:
            raster_transformers = [t.build() for t in self.transformers]
        else:
            raster_transformers = []

        built_raster_sources = [
            rs.build(tmp_dir, use_transformers) for rs in self.raster_sources
        ]
        multi_raster_source = MultiRasterSource(
            raster_sources=built_raster_sources,
            primary_source_idx=self.primary_source_idx,
            force_same_dtype=self.force_same_dtype,
            channel_order=self.channel_order,
            raster_transformers=raster_transformers,
            extent=self.extent)
        return multi_raster_source

    def update(self, pipeline=None, scene=None):
        for t in self.transformers:
            t.update(pipeline, scene)



from rastervision.core.data.label_source import (LabelSourceConfig,
                                                 ObjectDetectionLabelSource)
from rastervision.core.data.vector_source import VectorSourceConfig
from rastervision.core.data.vector_transformer import (
    ClassInferenceTransformerConfig, BufferTransformerConfig)
from rastervision.pipeline.config import register_config, validator


@register_config('object_detection_label_source')
class ObjectDetectionLabelSourceConfig(LabelSourceConfig):
    """Configure an :class:`.ObjectDetectionLabelSource`."""

    vector_source: VectorSourceConfig

    @validator('vector_source')
    def ensure_required_transformers(
            cls, v: VectorSourceConfig) -> VectorSourceConfig:
        """Add class-inference and buffer transformers if absent."""
        tfs = v.transformers

        # add class inference transformer
        has_inf_tf = any(
            isinstance(tf, ClassInferenceTransformerConfig) for tf in tfs)
        if not has_inf_tf:
            tfs += [ClassInferenceTransformerConfig(default_class_id=None)]

        # add buffer transformers
        has_buf_tf = any(isinstance(tf, BufferTransformerConfig) for tf in tfs)
        if not has_buf_tf:
            tfs += [
                BufferTransformerConfig(geom_type='Point', default_buf=1),
                BufferTransformerConfig(geom_type='LineString', default_buf=1)
            ]

        return v

    def update(self, pipeline=None, scene=None):
        super().update(pipeline, scene)
        self.vector_source.update(pipeline, scene)

    def build(self, class_config, crs_transformer, extent,
              tmp_dir=None) -> ObjectDetectionLabelSource:
        vs = self.vector_source.build(class_config, crs_transformer)
        return ObjectDetectionLabelSource(vs, extent)



from typing import List, Optional
import click

from rastervision.pipeline.file_system import get_tmp_dir
from rastervision.core.predictor import Predictor


# https://stackoverflow.com/questions/48391777/nargs-equivalent-for-options-in-click
class OptionEatAll(click.Option):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):
        def is_next_option(arg: str) -> bool:
            for prefix in self._eat_all_parser.prefixes:
                if arg.startswith(prefix):
                    return True
            return False

        def parser_process(value, state):
            # method to hook to the parser.process
            values = [value]
            # grab everything up to the next option
            while state.rargs and not is_next_option(state.rargs[0]):
                values.append(state.rargs.pop(0))
            # call the actual process
            self._previous_parser_process(values, state)

        retval = super().add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = (parser._long_opt.get(name)
                          or parser._short_opt.get(name))
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break

        return retval


@click.command(
    'predict', short_help='Use a model bundle to predict on new images.')
@click.argument('model_bundle')
@click.argument('image_uri')
@click.argument('label_uri')
@click.option(
    '--update-stats',
    '-a',
    is_flag=True,
    help=('Run an analysis on this individual image, as '
          'opposed to using any analysis like statistics '
          'that exist in the prediction package'))
@click.option(
    '--channel-order',
    cls=OptionEatAll,
    # https://stackoverflow.com/questions/48391777/nargs-equivalent-for-options-in-click#comment121399899_48394004
    type=list,
    help='List of indices comprising channel_order. Example: 2 1 0')
@click.option(
    '--scene-group',
    help='Name of the scene group whose stats will be used by the '
    'StatsTransformer. Requires the stats for this scene group to be present '
    'inside the bundle.')
def predict(model_bundle: str,
            image_uri: str,
            label_uri: str,
            update_stats: bool = False,
            channel_order: Optional[List[str]] = None,
            scene_group: Optional[str] = None):
    """Make predictions on the images at IMAGE_URI
    using MODEL_BUNDLE and store the prediction output at LABEL_URI.
    """
    if channel_order is not None:
        channel_order: List[int] = [int(i) for i in channel_order]

    with get_tmp_dir() as tmp_dir:
        predictor = Predictor(model_bundle, tmp_dir, update_stats,
                              channel_order, scene_group)
        predictor.predict([image_uri], label_uri)



from typing import (TYPE_CHECKING, Callable, Dict, Iterable, Iterator, List,
                    Optional, Union)
from copy import deepcopy

from shapely.geometry import shape, mapping
from tqdm.auto import tqdm

from rastervision.core.data.utils.misc import listify_uris

if TYPE_CHECKING:
    from rastervision.core.data.crs_transformer import CRSTransformer
    from shapely.geometry.base import BaseGeometry

MULTI_GEOM_TYPES = {'MultiPolygon', 'MultiPoint', 'MultiLineString'}
PROGRESSBAR_DELAY_SEC = 5


def geometry_to_feature(mapping: dict,
                        properties: Optional[dict] = None) -> dict:
    """Convert a serialized geometry to a serialized GeoJSON feature."""
    already_a_feature = mapping.get('type') == 'Feature'
    if already_a_feature:
        return mapping
    if properties is None:
        properties = {}
    return {'type': 'Feature', 'geometry': mapping, 'properties': properties}


def geometries_to_geojson(geometries: Iterable[dict]) -> dict:
    """Convert serialized geometries to a serialized GeoJSON FeatureCollection.
    """
    features = [geometry_to_feature(g) for g in geometries]
    return features_to_geojson(features)


def features_to_geojson(features: List[dict]) -> dict:
    """Convert GeoJSON-like mapping of Features to a FeatureCollection."""
    return {'type': 'FeatureCollection', 'features': features}


def map_features(func: Callable,
                 geojson: dict,
                 include_geom_types: Iterable[str] = [],
                 progressbar_kw: Optional[dict] = None) -> dict:
    """Map GeoJSON features to new features. Returns a new GeoJSON dict."""
    features_in = geojson['features']

    if progressbar_kw is None:
        progressbar_kw = dict(desc='Transforming features')

    with tqdm(
            features_in,
            delay=PROGRESSBAR_DELAY_SEC,
            mininterval=0.5,
            **progressbar_kw) as bar:
        if len(include_geom_types) > 0:
            include_geom_types = set(include_geom_types)
            features_out = [
                func(f)
                if f['geometry']['type'] in include_geom_types else deepcopy(f)
                for f in bar
            ]
        else:
            features_out = [func(f) for f in bar]

    return features_to_geojson(features_out)


def map_geoms(func: Callable,
              geojson: dict,
              include_geom_types: Iterable[str] = [],
              progressbar_kw: Optional[dict] = None) -> dict:
    """Map GeoJSON features to new features by applying func to geometries.

    Returns a new GeoJSON dict.
    """

    def feat_func(feature_in: dict) -> dict:
        # to shapely geometry
        geom_in = shape(feature_in['geometry'])
        # apply func
        geom_out = func(geom_in, feature=feature_in)
        # back to dict
        geom_out = mapping(geom_out)
        # new feature with updated geometry
        feature_out = geometry_to_feature(geom_out,
                                          feature_in.get('properties', {}))
        return feature_out

    if progressbar_kw is None:
        progressbar_kw = dict(desc='Transforming geoms')

    return map_features(
        feat_func,
        geojson,
        include_geom_types=include_geom_types,
        progressbar_kw=progressbar_kw)


def geojson_to_geoms(geojson: dict) -> Iterator['BaseGeometry']:
    """Return the shapely geometry for each feature in the GeoJSON."""
    geoms = (shape(f['geometry']) for f in geojson['features'])
    return geoms


def geoms_to_geojson(geoms: Iterable['BaseGeometry'],
                     properties: Optional[Iterable[dict]] = None) -> dict:
    """Serialize shapely geometries to GeoJSON."""
    if properties is None:
        features = [geom_to_feature(g) for g in geoms]
    else:
        features = [geom_to_feature(g, p) for g, p in zip(geoms, properties)]
    geojson = features_to_geojson(features)
    return geojson


def geom_to_feature(geom: 'BaseGeometry',
                    properties: Optional[dict] = None) -> dict:
    """Serialize a single shapely geomety to a GeoJSON Feature."""
    geomety = mapping(geom)
    feature = geometry_to_feature(geomety, properties=properties)
    return feature


def filter_features(func: Callable,
                    geojson: dict,
                    progressbar_kw: Optional[dict] = None) -> dict:
    """Filter GeoJSON features. Returns a new GeoJSON dict."""
    features_in = geojson['features']

    if progressbar_kw is None:
        progressbar_kw = dict(desc='Filtering features.')

    with tqdm(
            features_in,
            delay=PROGRESSBAR_DELAY_SEC,
            mininterval=0.5,
            **progressbar_kw) as bar:
        features_out = list(filter(func, bar))

    return features_to_geojson(features_out)


def is_empty_feature(f: dict) -> bool:
    """Check if a GeoJSON Feature lacks geometry info.

    This was added to handle empty geoms which appear when using
    OSM vector tiles.

    Args:
        f (dict): A GeoJSON-like mapping of a Feature.

    Returns:
        bool: Whether the feature contains any geometry.
    """
    g: Optional[dict] = f.get('geometry')
    if not g:
        return True
    no_geometries = not g.get('geometries')
    no_coordinates = not g.get('coordinates')
    is_empty = no_geometries and no_coordinates
    return is_empty


def remove_empty_features(geojson: dict) -> dict:
    """Remove Features from a FeatureCollection that lack geometry info.

    Args:
        geojson (dict): A GeoJSON-like mapping of a FeatureCollection.

    Returns:
        dict: Filtered FeatureCollection.
    """
    return filter_features(
        lambda f: not is_empty_feature(f),
        geojson,
        progressbar_kw=dict(desc='Removing empty features'))


def split_multi_geometries(geojson: dict) -> dict:
    """Split any Features with multi-part geometries into multiple Features.

    Args:
        geojson (dict): A GeoJSON-like mapping of a FeatureCollection.

    Returns:
        dict: FeatureCollection without multi-part geometries.
    """

    def split_geom(geom: 'BaseGeometry') -> List['BaseGeometry']:
        # Split GeometryCollection into list of geoms.
        if geom.geom_type == 'GeometryCollection':
            geoms = list(geom)
        else:
            geoms = [geom]
        # Split any MultiX to list of X.
        new_geoms = []
        for g in geoms:
            if g.geom_type in MULTI_GEOM_TYPES:
                new_geoms.extend(list(g.geoms))
            else:
                new_geoms.append(g)
        return new_geoms

    include_geom_types = set(['GeometryCollection', *MULTI_GEOM_TYPES])

    all_geom_types = set(f['geometry']['type'] for f in geojson['features'])
    if len(include_geom_types.intersection(all_geom_types)) == 0:
        return geojson

    new_features = []
    with tqdm(
            geojson['features'],
            desc='Splitting multi-part geoms',
            delay=PROGRESSBAR_DELAY_SEC,
            mininterval=0.5) as bar:
        for f in bar:
            geom_type = f['geometry']['type']
            if geom_type not in include_geom_types:
                new_features.append(deepcopy(f))
                continue
            geom = shape(f['geometry'])
            split_geoms = split_geom(geom)
            for g in split_geoms:
                new_feature = geometry_to_feature(
                    mapping(g), f.get('properties', {}))
                new_features.append(new_feature)
    return features_to_geojson(new_features)


def map_to_pixel_coords(geojson: dict,
                        crs_transformer: 'CRSTransformer') -> dict:
    """Convert a GeoJSON dict from map to pixel coordinates."""
    return map_geoms(
        lambda g, **kw: crs_transformer.map_to_pixel(g),
        geojson,
        progressbar_kw=dict(desc='Transforming to pixel coords'))


def pixel_to_map_coords(geojson: dict,
                        crs_transformer: 'CRSTransformer') -> dict:
    """Convert a GeoJSON dict from pixel to map coordinates."""
    return map_geoms(
        lambda g, **kw: crs_transformer.pixel_to_map(g),
        geojson,
        progressbar_kw=dict(desc='Transforming to map coords'))


def simplify_polygons(geojson: dict) -> dict:
    """Simplify polygon geometries by applying ``.buffer(0)``.

    For Polygon geomtries, ``.buffer(0)`` can do the following:

    1.  *Sometimes* break up a polygon with "bowties" into multiple polygons.
        (See https://github.com/shapely/shapely/issues/599.)
    2.  *Sometimes* "simplify" polygons. See shapely documentation for
        :meth:`.BaseGeometry.buffer`.

    Args:
        geojson (dict): A GeoJSON-like mapping of a FeatureCollection.

    Returns:
        dict: FeatureCollection with simplified geometries.
    """

    all_geom_types = set(f['geometry']['type'] for f in geojson['features'])
    if 'Polygon' not in all_geom_types:
        return geojson

    geojson_buffered = map_geoms(
        lambda g, **kw: g.buffer(0),
        geojson,
        include_geom_types=['Polygon'],
        progressbar_kw=dict(desc='Simplifying polygons'))
    geojson_cleaned = remove_empty_features(geojson_buffered)
    geojson_split = split_multi_geometries(geojson_cleaned)
    return geojson_split


def buffer_geoms(geojson: dict,
                 geom_type: str,
                 class_bufs: Dict[int, Optional[float]] = {},
                 default_buf: Optional[float] = 1) -> dict:
    """Buffer geometries.

    Geometries in features without a class_id property will be ignored.

    Args:
        geojson (dict): A GeoJSON-like mapping of a FeatureCollection.
        geom_type (str): Shapely geometry type to apply the buffering to. Other
            types of geometries will not be affected.
        class_bufs (Dict[int, Optional[float]]): Optional
            mapping from class ID to buffer distance (in pixel units) for
            geom_type geometries.

    Returns:
        dict: FeatureCollection with buffered geometries.
    """

    def buffer_geom(geom: 'BaseGeometry',
                    feature: Optional[dict] = None) -> 'BaseGeometry':
        has_class_id = (('properties' in feature)
                        and ('class_id' in feature['properties']))
        if has_class_id:
            class_id = feature['properties']['class_id']
            buf = class_bufs.get(class_id, default_buf)
        else:
            buf = default_buf

        # If buf for the class_id was explicitly set as None, don't buffer.
        if buf is not None:
            geom = geom.buffer(buf)

        return geom

    all_geom_types = set(f['geometry']['type'] for f in geojson['features'])
    if geom_type not in all_geom_types:
        return geojson

    geojson_buffered = map_geoms(
        buffer_geom,
        geojson,
        include_geom_types=[geom_type],
        progressbar_kw=dict(desc=f'Buffering {geom_type}s (if any)'))
    return geojson_buffered


def all_geoms_valid(geojson: dict):
    """Check if there are any invalid geometries in the GeoJSON."""
    geoms = geojson_to_geoms(geojson)
    return all(g.is_valid for g in geoms)


def get_polygons_from_uris(
        uris: Union[str, List[str]],
        crs_transformer: 'CRSTransformer') -> List['BaseGeometry']:
    """Load and return polygons (in pixel coords) from one or more URIs."""

    # use local imports to avoid circular import problems
    from rastervision.core.data import GeoJSONVectorSource

    polygons = []
    uris = listify_uris(uris)
    for uri in uris:
        source = GeoJSONVectorSource(
            uri=uri, ignore_crs_field=True, crs_transformer=crs_transformer)
        polygons += source.get_geoms()
    return polygons



# flake8: noqa

from rastervision.core.data.utils.misc import *
from rastervision.core.data.utils.geojson import *
from rastervision.core.data.utils.factory import *
from rastervision.core.data.utils.vectorization import *



from typing import TYPE_CHECKING, List

from rastervision.core.data.vector_source.vector_source import VectorSource
from rastervision.pipeline.file_system import download_if_needed, file_to_json

if TYPE_CHECKING:
    from rastervision.core.data import CRSTransformer, VectorTransformer


class GeoJSONVectorSource(VectorSource):
    """A :class:`.VectorSource` for reading GeoJSON files."""

    def __init__(self,
                 uri: str,
                 crs_transformer: 'CRSTransformer',
                 vector_transformers: List['VectorTransformer'] = [],
                 ignore_crs_field: bool = False):
        """Constructor.

        Args:
            uri (str): URI of the GeoJSON file.
            crs_transformer: A ``CRSTransformer`` to convert
                between map and pixel coords. Normally this is obtained from a
                :class:`.RasterSource`.
            vector_transformers: ``VectorTransformers`` for transforming
                geometries. Defaults to ``[]``.
            ignore_crs_field (bool): Ignore the CRS specified in the file and
                assume WGS84 (EPSG:4326) coords. Only WGS84 is supported
                currently. If False, and the file contains a CRS, will throw an
                exception on read. Defaults to False.
        """
        self.uri = uri
        self.ignore_crs_field = ignore_crs_field
        super().__init__(
            crs_transformer, vector_transformers=vector_transformers)

    def _get_geojson(self):
        # download first so that it gets cached
        geojson = file_to_json(download_if_needed(self.uri))
        if not self.ignore_crs_field and 'crs' in geojson:
            raise NotImplementedError(
                f'The GeoJSON file at {self.uri} contains a CRS field which '
                'is not allowed by the current GeoJSON standard or by '
                'Raster Vision. All coordinates are expected to be in '
                'EPSG:4326 CRS. If the file uses EPSG:4326 (ie. lat/lng on '
                'the WGS84 reference ellipsoid) and you would like to ignore '
                'the CRS field, set ignore_crs_field=True in '
                'GeoJSONVectorSourceConfig.')

        return geojson



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Optional, overload, Tuple

from rasterio.windows import Window
from shapely.ops import transform
from shapely.geometry.base import BaseGeometry

from rastervision.core.box import Box

if TYPE_CHECKING:
    import numpy as np


class CRSTransformer(ABC):
    """Transforms map points in some CRS into pixel coordinates.

    Each transformer is associated with a particular :class:`.RasterSource`.
    """

    def __init__(self,
                 transform: Optional[Any] = None,
                 image_crs: Optional[str] = None,
                 map_crs: Optional[str] = None):
        self.transform = transform
        self.image_crs = image_crs
        self.map_crs = map_crs

    @overload
    def map_to_pixel(self, inp: Tuple[float, float]) -> Tuple[int, int]:
        ...

    @overload
    def map_to_pixel(self, inp: Tuple['np.array', 'np.array']
                     ) -> Tuple['np.array', 'np.array']:
        ...

    @overload
    def map_to_pixel(self, inp: Box) -> Box:
        ...

    @overload
    def map_to_pixel(self, inp: BaseGeometry) -> BaseGeometry:
        ...

    def map_to_pixel(self, inp):
        """Transform input from pixel to map coords.

        Args:
            inp: (x, y) tuple or Box or rasterio Window or shapely geometry in
                pixel coordinates. If tuple, x and y can be single values or
                array-like.

        Returns:
            Coordinate-transformed input in the same format.
        """
        if isinstance(inp, Box):
            box_in = inp
            ymin, xmin, ymax, xmax = box_in
            xmin_tf, ymin_tf = self._map_to_pixel((xmin, ymin))
            xmax_tf, ymax_tf = self._map_to_pixel((xmax, ymax))
            box_out = Box(ymin_tf, xmin_tf, ymax_tf, xmax_tf)
            return box_out
        elif isinstance(inp, Window):
            window_in = inp
            (ymin, ymax), (xmin, xmax) = window_in.toranges()
            xmin_tf, ymin_tf = self._map_to_pixel((xmin, ymin))
            xmax_tf, ymax_tf = self._map_to_pixel((xmax, ymax))
            window_out = Window.from_slices(
                slice(ymin_tf, ymax_tf), slice(xmin_tf, xmax_tf))
            return window_out
        elif isinstance(inp, BaseGeometry):
            geom_in = inp
            geom_out = transform(
                lambda x, y, z=None: self._map_to_pixel((x, y)), geom_in)
            return geom_out
        elif len(inp) == 2:
            return self._map_to_pixel(inp)
        else:
            raise TypeError('Input must be 2-tuple or Box or rasterio Window '
                            'or shapely geometry.')

    @overload
    def pixel_to_map(self, inp: Tuple[float, float]) -> Tuple[float, float]:
        ...

    @overload
    def pixel_to_map(self, inp: Tuple['np.array', 'np.array']
                     ) -> Tuple['np.array', 'np.array']:
        ...

    @overload
    def pixel_to_map(self, inp: Box) -> Box:
        ...

    @overload
    def pixel_to_map(self, inp: BaseGeometry) -> BaseGeometry:
        ...

    def pixel_to_map(self, inp):
        """Transform input from pixel to map coords.

        Args:
            inp: (x, y) tuple or Box or rasterio Window or shapely geometry in
                pixel coordinates. If tuple, x and y can be single values or
                array-like.

        Returns:
            Coordinate-transformed input in the same format.
        """
        if isinstance(inp, Box):
            box_in = inp
            ymin, xmin, ymax, xmax = box_in
            xmin_tf, ymin_tf = self._pixel_to_map((xmin, ymin))
            xmax_tf, ymax_tf = self._pixel_to_map((xmax, ymax))
            box_out = Box(ymin_tf, xmin_tf, ymax_tf, xmax_tf)
            return box_out
        elif isinstance(inp, Window):
            window_in = inp
            (ymin, ymax), (xmin, xmax) = window_in.toranges()
            xmin_tf, ymin_tf = self._pixel_to_map((xmin, ymin))
            xmax_tf, ymax_tf = self._pixel_to_map((xmax, ymax))
            window_out = Window.from_slices(
                slice(ymin_tf, ymax_tf), slice(xmin_tf, xmax_tf))
            return window_out
        elif isinstance(inp, BaseGeometry):
            geom_in = inp
            geom_out = transform(
                lambda x, y, z=None: self._pixel_to_map((x, y)), geom_in)
            return geom_out
        elif len(inp) == 2:
            return self._pixel_to_map(inp)
        else:
            raise TypeError('Input must be 2-tuple or Box or rasterio Window '
                            'or shapely geometry.')

    @abstractmethod
    def _map_to_pixel(self, point: Tuple[float, float]) -> Tuple[int, int]:
        """Transform point from map to pixel coordinates.

        Args:
            map_point: (x, y) tuple in map coordinates (eg. lon/lat). x and y
            can be single values or array-like.

        Returns:
            Tuple[int, int]: (x, y) tuple in pixel coordinates.
        """
        pass

    @abstractmethod
    def _pixel_to_map(self, point: Tuple[float, float]) -> Tuple[float, float]:
        """Transform point from pixel to map coordinates.

        Args:
            pixel_point: (x, y) tuple in pixel coordinates. x and y can be
            single values or array-like.

        Returns:
            Tuple[float, float]: (x, y) tuple in map coordinates (eg. lon/lat).
        """
        pass



from typing import Optional, List, TYPE_CHECKING

from rastervision.pipeline.config import Config, register_config

if TYPE_CHECKING:
    from rastervision.core.rv_pipeline.rv_pipeline import RVPipeline  # noqa
    from rastervision.core.backend.backend import Backend  # noqa


@register_config('backend')
class BackendConfig(Config):
    """Configure a :class:`.Backend`."""

    def build(self, pipeline: 'RVPipeline', tmp_dir: str) -> 'Backend':
        raise NotImplementedError()

    def get_bundle_filenames(self) -> List[str]:
        """Returns the names of files that should be included in a model bundle.

        The files are assumed to be in the train/ directory generated by the train
        command. Note that only the names, not the full paths should be returned.
        """
        raise NotImplementedError()

    def update(self, pipeline: Optional['RVPipeline'] = None):  # noqa
        pass

    def filter_commands(self, commands: List[str]) -> List[str]:
        """Filter out any commands that are not needed or supported."""
        return commands



from typing import List, Union

from rastervision.core.data.raster_source import RasterSourceConfig, RasterioSource
from rastervision.pipeline.config import ConfigError, Field, register_config


def rasterio_source_config_upgrader(cfg_dict: dict, version: int) -> dict:
    if version == 5:
        # removed in version 6
        x_shift = cfg_dict.get('x_shift', 0)
        y_shift = cfg_dict.get('y_shift', 0)
        if x_shift != 0 or y_shift != 0:
            raise ConfigError('x_shift and y_shift are deprecated. '
                              'Use the ShiftTrasnformer instead.')
        try:
            del cfg_dict['x_shift']
            del cfg_dict['y_shift']
        except KeyError:
            pass
    return cfg_dict


@register_config('rasterio_source', upgrader=rasterio_source_config_upgrader)
class RasterioSourceConfig(RasterSourceConfig):
    """Configure a :class:`.RasterioSource`."""

    uris: Union[str, List[str]] = Field(
        ...,
        description='One or more image URIs that comprise the imagery for a '
        'scene. The format of each file can be any that can be read by '
        'Rasterio/GDAL. If > 1 URI is provided, a VRT will be created to '
        'mosaic together the individual images.')
    allow_streaming: bool = Field(
        False,
        description='Stream assets as needed rather than downloading them.')

    def build(self, tmp_dir, use_transformers=True):
        raster_transformers = ([rt.build() for rt in self.transformers]
                               if use_transformers else [])

        return RasterioSource(
            uris=self.uris,
            raster_transformers=raster_transformers,
            tmp_dir=tmp_dir,
            allow_streaming=self.allow_streaming,
            channel_order=self.channel_order,
            extent=self.extent)



from typing import List
from abc import (ABC, abstractmethod)

from rastervision.core.data import Scene


class Analyzer(ABC):
    """Analyzes scenes and writes some output while running the analyze command.

    This output can be used to normalize images, for example.
    """

    @abstractmethod
    def process(self, scenes: List[Scene], tmp_dir: str):
        """Process scenes and save result."""
        pass



# flake8: noqa

from rastervision.core.data.raster_transformer.raster_transformer import *
from rastervision.core.data.raster_transformer.raster_transformer_config import *
from rastervision.core.data.raster_transformer.stats_transformer import *
from rastervision.core.data.raster_transformer.stats_transformer_config import *
from rastervision.core.data.raster_transformer.nan_transformer import *
from rastervision.core.data.raster_transformer.nan_transformer_config import *
from rastervision.core.data.raster_transformer.cast_transformer import *
from rastervision.core.data.raster_transformer.cast_transformer_config import *
from rastervision.core.data.raster_transformer.reclass_transformer import *
from rastervision.core.data.raster_transformer.reclass_transformer_config import *
from rastervision.core.data.raster_transformer.min_max_transformer import *
from rastervision.core.data.raster_transformer.min_max_transformer_config import *
from rastervision.core.data.raster_transformer.rgb_class_transformer import *
from rastervision.core.data.raster_transformer.rgb_class_transformer_config import *

__all__ = [
    RasterTransformer.__name__,
    RasterTransformerConfig.__name__,
    StatsTransformer.__name__,
    StatsTransformerConfig.__name__,
    NanTransformer.__name__,
    NanTransformerConfig.__name__,
    CastTransformer.__name__,
    CastTransformerConfig.__name__,
    ReclassTransformer.__name__,
    ReclassTransformerConfig.__name__,
    MinMaxTransformer.__name__,
    MinMaxTransformerConfig.__name__,
    RGBClassTransformer.__name__,
    RGBClassTransformerConfig.__name__,
]



from typing import Optional, List

from rastervision.pipeline.config import (Config, ConfigError, register_config,
                                          Field)
from rastervision.core.data.raster_source import RasterSourceConfig
from rastervision.core.data.label_source import LabelSourceConfig
from rastervision.core.data.label_store import LabelStoreConfig
from rastervision.core.data.scene import Scene
from rastervision.core.data.utils import get_polygons_from_uris


def scene_config_upgrader(cfg_dict: dict, version: int) -> dict:
    if version == 4:
        try:
            # removed in version 5
            if cfg_dict.get('aoi_geometries') is not None:
                raise ConfigError(
                    'SceneConfig.aoi_geometries is deprecated. '
                    'To use this config again, manually edit it to use '
                    'SceneConfig.aoi_uris instead.')
            del cfg_dict['aoi_geometries']
        except KeyError:
            pass
    return cfg_dict


@register_config('scene', upgrader=scene_config_upgrader)
class SceneConfig(Config):
    """Configure a :class:`.Scene` comprising raster data & labels for an AOI.
    """

    id: str
    raster_source: RasterSourceConfig
    label_source: Optional[LabelSourceConfig] = None
    label_store: Optional[LabelStoreConfig] = None
    aoi_uris: Optional[List[str]] = Field(
        None,
        description='List of URIs of GeoJSON files that define the AOIs for '
        'the scene. Each polygon defines an AOI which is a piece of the scene '
        'that is assumed to be fully labeled and usable for training or '
        'validation. The AOIs are assumed to be in EPSG:4326 coordinates.')

    def build(self, class_config, tmp_dir, use_transformers=True) -> Scene:
        raster_source = self.raster_source.build(
            tmp_dir, use_transformers=use_transformers)
        crs_transformer = raster_source.crs_transformer
        extent = raster_source.extent

        label_source = (self.label_source.build(class_config, crs_transformer,
                                                extent, tmp_dir)
                        if self.label_source is not None else None)
        label_store = (self.label_store.build(class_config, crs_transformer,
                                              extent, tmp_dir)
                       if self.label_store is not None else None)

        aoi_polygons = []
        if self.aoi_uris is not None:
            aoi_polygons += get_polygons_from_uris(self.aoi_uris,
                                                   crs_transformer)

        return Scene(
            self.id,
            raster_source,
            label_source=label_source,
            label_store=label_store,
            aoi_polygons=aoi_polygons)

    def update(self, pipeline=None):
        super().update()

        self.raster_source.update(pipeline=pipeline, scene=self)
        if self.label_source is not None:
            self.label_source.update(pipeline=pipeline, scene=self)
        if self.label_store is None and pipeline is not None:
            self.label_store = pipeline.get_default_label_store(scene=self)
        if self.label_store is not None:
            self.label_store.update(pipeline=pipeline, scene=self)



from typing import TYPE_CHECKING, List, Optional

import numpy as np

from rastervision.core.data.raster_transformer import (RasterTransformer,
                                                       ReclassTransformer)
from rastervision.core.data.utils import (color_to_triple, color_to_integer,
                                          rgb_to_int_array)

if TYPE_CHECKING:
    from rastervision.core.data.class_config import ClassConfig


class RGBClassTransformer(RasterTransformer):
    """Maps RGB values to class IDs. Can also do the reverse."""

    def __init__(self, class_config: 'ClassConfig'):
        class_config.ensure_null_class()
        self.null_class_id = class_config.null_class_id
        color_to_class = class_config.get_color_to_class_id()

        self.rgb_int_to_class = {
            color_to_integer(col): class_id
            for col, class_id in color_to_class.items()
        }
        self.rgb_int_to_class_tf = ReclassTransformer(self.rgb_int_to_class)

        class_to_color_triple = {
            class_id: color_to_triple(col)
            for col, class_id in color_to_class.items()
        }
        # i-th row of this array is the color-triple of the i-th class
        self.class_to_rgb_arr = np.array(
            [
                class_to_color_triple[c]
                for c in sorted(class_to_color_triple.keys())
            ],
            dtype=np.uint8)

    def transform(self,
                  chip: np.ndarray,
                  channel_order: Optional[List[int]] = None) -> np.ndarray:
        """Transform RGB array to array of class IDs or vice versa.

        Args:
            chip (np.ndarray): Numpy array of shape (H, W, 3).
            channel_order (Optional[List[int]], optional): List of indices of
                channels that were extracted from the raw imagery.
                Defaults to None.

        Returns:
            np.ndarray: An array of class IDs.
        """
        return self.rgb_to_class(chip)

    def rgb_to_class(self, array_rgb: np.ndarray) -> np.ndarray:
        array_int = rgb_to_int_array(array_rgb)
        array_class_id = self.rgb_int_to_class_tf.transform(array_int)
        return array_class_id.astype(np.uint8)

    def class_to_rgb(self, class_labels: np.ndarray) -> np.ndarray:
        return self.class_to_rgb_arr[class_labels]



# flake8: noqa

from rastervision.core.data.crs_transformer.crs_transformer import *
from rastervision.core.data.crs_transformer.identity_crs_transformer import *
from rastervision.core.data.crs_transformer.rasterio_crs_transformer import *

__all__ = [
    CRSTransformer.__name__,
    RasterioCRSTransformer.__name__,
    IdentityCRSTransformer.__name__,
]



from typing import (TYPE_CHECKING, Iterable, Iterator, List, Tuple)
if TYPE_CHECKING:
    import numpy as np
    from rastervision.core.box import Box


def discard_prediction_edges(
        windows: Iterable['Box'], predictions: Iterable['np.ndarray'],
        crop_sz: int) -> Tuple[List['Box'], Iterator['np.ndarray']]:
    """Discard the edges of predicted chips.

    Args:
        windows (Iterable[Box]): The windows corresponding to the chips.
        predictions (Iterable[np.ndarray]): The predicted chips.
        crop_sz (int): Number of pixel rows/cols to discard.

    Returns:
        Tuple[Iterator[Box], Iterator[np.ndarray]]: Cropped windows and chips.
    """
    windows_cropped = [w.center_crop(crop_sz, crop_sz) for w in windows]
    array_slices = [
        wc.to_offsets(w).to_slices() for w, wc in zip(windows, windows_cropped)
    ]
    predictions_cropped = (p[..., yslice, xslice]
                           for p, (xslice,
                                   yslice) in zip(predictions, array_slices))
    return windows_cropped, predictions_cropped



# flake8: noqa


def register_plugin(registry):
    registry.set_plugin_version('rastervision.core', 7)
    from rastervision.core.cli import predict
    registry.add_plugin_command(predict)


import rastervision.pipeline
from rastervision.core.box import *
from rastervision.core.data_sample import *
from rastervision.core.predictor import *
from rastervision.core.raster_stats import *

# We just need to import anything that contains a Config, so that all
# the register_config decorators will be called which add Configs to the
# registry.
import rastervision.core.backend
import rastervision.core.data
import rastervision.core.rv_pipeline
import rastervision.core.evaluation



from typing import TYPE_CHECKING, Callable, Dict, Union, Tuple, Optional, List
from typing_extensions import Literal
from pydantic import PositiveInt as PosInt, conint
import math
import random

import numpy as np
from shapely.geometry import Polygon
from rasterio.windows import Window as RioWindow

NonNegInt = conint(ge=0)

if TYPE_CHECKING:
    pass


class BoxSizeError(ValueError):
    pass


class Box():
    """A multi-purpose box (ie. rectangle) representation ."""

    def __init__(self, ymin, xmin, ymax, xmax):
        """Construct a bounding box.

        Unless otherwise stated, the convention is that these coordinates are
        in pixel coordinates and represent boxes that lie within a
        RasterSource.

        Args:
            ymin: minimum y value (y is row)
            xmin: minimum x value (x is column)
            ymax: maximum y value
            xmax: maximum x value

        """
        ymin, ymax = sorted((ymin, ymax))
        xmin, xmax = sorted((xmin, xmax))
        self.ymin = ymin
        self.xmin = xmin
        self.ymax = ymax
        self.xmax = xmax

    def __eq__(self, other: 'Box') -> bool:
        """Return true if other has same coordinates."""
        return self.tuple_format() == other.tuple_format()

    def __ne__(self, other: 'Box'):
        """Return true if other has different coordinates."""
        return self.tuple_format() != other.tuple_format()

    @property
    def height(self) -> int:
        """Return height of Box."""
        return self.ymax - self.ymin

    @property
    def width(self) -> int:
        """Return width of Box."""
        return self.xmax - self.xmin

    @property
    def size(self) -> Tuple[int, int]:
        return self.height, self.width

    @property
    def area(self) -> int:
        """Return area of Box."""
        return self.height * self.width

    def rasterio_format(self) -> Tuple[Tuple[int, int], Tuple[int, int]]:
        """Return Box in Rasterio format: ((ymin, ymax), (xmin, xmax))."""
        return ((self.ymin, self.ymax), (self.xmin, self.xmax))

    def tuple_format(self) -> Tuple[int, int, int, int]:
        return (self.ymin, self.xmin, self.ymax, self.xmax)

    def shapely_format(self) -> Tuple[int, int, int, int]:
        return self.to_xyxy()

    def to_int(self):
        return Box(
            int(self.ymin), int(self.xmin), int(self.ymax), int(self.xmax))

    def npbox_format(self):
        """Return Box in npbox format used by TF Object Detection API.

        Returns:
            Numpy array of form [ymin, xmin, ymax, xmax] with float type

        """
        return np.array(
            [self.ymin, self.xmin, self.ymax, self.xmax], dtype=float)

    @staticmethod
    def to_npboxes(boxes):
        """Return nx4 numpy array from list of Box."""
        nb_boxes = len(boxes)
        npboxes = np.empty((nb_boxes, 4))
        for boxind, box in enumerate(boxes):
            npboxes[boxind, :] = box.npbox_format()
        return npboxes

    def __iter__(self):
        return iter(self.tuple_format())

    def __getitem__(self, i):
        return self.tuple_format()[i]

    def __repr__(self) -> str:
        arg_keys = ['ymin', 'xmin', 'ymax', 'xmax']
        arg_vals = [getattr(self, k) for k in arg_keys]
        arg_strs = [f'{k}={v}' for k, v in zip(arg_keys, arg_vals)]
        arg_str = ', '.join(arg_strs)
        return f'{type(self).__name__}({arg_str})'

    def __hash__(self) -> int:
        return hash(self.tuple_format())

    def geojson_coordinates(self) -> List[Tuple[int, int]]:
        """Return Box as GeoJSON coordinates."""
        # Compass directions:
        nw = [self.xmin, self.ymin]
        ne = [self.xmin, self.ymax]
        se = [self.xmax, self.ymax]
        sw = [self.xmax, self.ymin]
        return [nw, ne, se, sw, nw]

    def make_random_square_container(self, size):
        """Return a new square Box that contains this Box.

        Args:
            size: the width and height of the new Box

        """
        if size < self.width:
            raise BoxSizeError('size of random container cannot be < width')

        if size < self.height:  # pragma: no cover
            raise BoxSizeError('size of random container cannot be < height')

        lb = self.ymin - (size - self.height)
        ub = self.ymin
        rand_y = random.randint(int(lb), int(ub))

        lb = self.xmin - (size - self.width)
        ub = self.xmin
        rand_x = random.randint(int(lb), int(ub))

        return Box.make_square(rand_y, rand_x, size)

    def make_random_box_container(self, out_h: int, out_w: int) -> 'Box':
        """Return a new rectangular Box that contains this Box.

        Args:
            out_h (int): the height of the new Box
            out_w (int): the width of the new Box
        """
        self_h, self_w = self.size

        if out_h < self_h:  # pragma: no cover
            raise BoxSizeError('size of random container cannot be < height')
        if out_w < self_w:
            raise BoxSizeError('size of random container cannot be < width')

        lb = self.ymin - (out_h - self_h)
        ub = self.ymin
        ymin = random.randint(int(lb), int(ub))

        lb = self.xmin - (out_w - self_w)
        ub = self.xmin
        xmin = random.randint(int(lb), int(ub))

        return Box(ymin, xmin, ymin + out_h, xmin + out_w)

    def make_random_square(self, size: int) -> 'Box':
        """Return new randomly positioned square Box that lies inside this Box.

        Args:
            size: the height and width of the new Box

        """
        if size >= self.width:
            raise BoxSizeError('size of random square cannot be >= width')

        if size >= self.height:  # pragma: no cover
            raise BoxSizeError('size of random square cannot be >= height')

        lb = self.ymin
        ub = self.ymax - size
        rand_y = random.randint(int(lb), int(ub))

        lb = self.xmin
        ub = self.xmax - size
        rand_x = random.randint(int(lb), int(ub))

        return Box.make_square(rand_y, rand_x, size)

    def intersection(self, other: 'Box') -> 'Box':
        """Return the intersection of this Box and the other.

        Args:
            other: The box to intersect with this one.

        Returns:
             The intersection of this box and the other one.

        """
        if not self.intersects(other):
            return Box(0, 0, 0, 0)
        xmin = max(self.xmin, other.xmin)
        ymin = max(self.ymin, other.ymin)
        xmax = min(self.xmax, other.xmax)
        ymax = min(self.ymax, other.ymax)
        return Box(xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)

    def intersects(self, other: 'Box') -> bool:
        if self.ymax <= other.ymin or self.ymin >= other.ymax:
            return False
        if self.xmax <= other.xmin or self.xmin >= other.xmax:
            return False
        return True

    @staticmethod
    def from_npbox(npbox):
        """Return new Box based on npbox format.

        Args:
            npbox: Numpy array of form [ymin, xmin, ymax, xmax] with float type

        """
        return Box(*npbox)

    @staticmethod
    def from_shapely(shape):
        xmin, ymin, xmax, ymax = shape.bounds
        return Box(ymin, xmin, ymax, xmax)

    @classmethod
    def from_rasterio(self, rio_window: RioWindow) -> 'Box':
        yslice, xslice = rio_window.toslices()
        return Box(yslice.start, xslice.start, yslice.stop, xslice.stop)

    def to_xywh(self) -> Tuple[int, int, int, int]:
        return (self.xmin, self.ymin, self.width, self.height)

    def to_xyxy(self) -> Tuple[int, int, int, int]:
        return (self.xmin, self.ymin, self.xmax, self.ymax)

    def to_points(self) -> np.ndarray:
        """Get (x, y) coords of each vertex as a 4x2 numpy array."""
        return np.array(self.geojson_coordinates()[:4])

    def to_shapely(self) -> Polygon:
        """Convert to shapely Polygon."""
        return Polygon.from_bounds(*(self.shapely_format()))

    def to_rasterio(self) -> RioWindow:
        """Convert to a Rasterio Window."""
        return RioWindow.from_slices(*self.to_slices())

    def to_slices(self) -> Tuple[slice, slice]:
        """Convert to slices: ymin:ymax, xmin:xmax"""
        return slice(self.ymin, self.ymax), slice(self.xmin, self.xmax)

    def translate(self, dy: int, dx: int) -> 'Box':
        """Translate window along y and x axes by the given distances."""
        ymin, xmin, ymax, xmax = self
        return Box(ymin + dy, xmin + dx, ymax + dy, xmax + dx)

    def shift_origin(self, extent: 'Box') -> 'Box':
        """Shift origin of window coords to (extent.xmin, extent.ymin)."""
        return self.translate(dy=extent.ymin, dx=extent.xmin)

    def to_offsets(self, container: 'Box') -> 'Box':
        """Convert coords to offsets from (container.xmin, container.ymin)."""
        return self.translate(dy=-container.ymin, dx=-container.xmin)

    def reproject(self, transform_fn: Callable) -> 'Box':
        """Reprojects this box based on a transform function.

        Args:
          transform_fn - A function that takes in a tuple (x, y)
                         and reprojects that point to the target
                         coordinate reference system.
        """
        (xmin, ymin) = transform_fn((self.xmin, self.ymin))
        (xmax, ymax) = transform_fn((self.xmax, self.ymax))

        return Box(ymin, xmin, ymax, xmax)

    @staticmethod
    def make_square(ymin, xmin, size) -> 'Box':
        """Return new square Box."""
        return Box(ymin, xmin, ymin + size, xmin + size)

    def center_crop(self, edge_offset_y: int, edge_offset_x: int) -> 'Box':
        """Return Box whose sides are eroded by the given offsets.

        Box(0, 0, 10, 10).center_crop(2, 4) ==  Box(2, 4, 8, 6)
        """
        return Box(self.ymin + edge_offset_y, self.xmin + edge_offset_x,
                   self.ymax - edge_offset_y, self.xmax - edge_offset_x)

    def erode(self, erosion_sz) -> 'Box':
        """Return new Box whose sides are eroded by erosion_sz."""
        return self.center_crop(erosion_sz, erosion_sz)

    def buffer(self, buffer_sz: float, max_extent: 'Box') -> 'Box':
        """Return new Box whose sides are buffered by buffer_sz.

        The resulting box is clipped so that the values of the corners are
        always greater than zero and less than the height and width of
        max_extent.

        """
        buffer_sz = max(0., buffer_sz)
        if buffer_sz < 1.:
            delta_width = int(round(buffer_sz * self.width))
            delta_height = int(round(buffer_sz * self.height))
        else:
            delta_height = delta_width = int(round(buffer_sz))

        return Box(
            max(0, math.floor(self.ymin - delta_height)),
            max(0, math.floor(self.xmin - delta_width)),
            min(max_extent.height,
                int(self.ymax) + delta_height),
            min(max_extent.width,
                int(self.xmax) + delta_width))

    def pad(self, ymin: int, xmin: int, ymax: int, xmax: int) -> 'Box':
        """Pad sides by the given amount."""
        return Box(
            ymin=self.ymin - ymin,
            xmin=self.xmin - xmin,
            ymax=self.ymax + ymax,
            xmax=self.xmax + xmax)

    def copy(self) -> 'Box':
        return Box(*self)

    def get_windows(self,
                    size: Union[PosInt, Tuple[PosInt, PosInt]],
                    stride: Union[PosInt, Tuple[PosInt, PosInt]],
                    padding: Optional[Union[NonNegInt, Tuple[
                        NonNegInt, NonNegInt]]] = None,
                    pad_direction: Literal['both', 'start', 'end'] = 'end'
                    ) -> List['Box']:
        """Returns a list of boxes representing windows generated using a
        sliding window traversal with the specified size, stride, and
        padding.

        Each of size, stride, and padding can be either a positive int or
        a tuple `(vertical-componet, horizontal-component)` of positive ints.

        Padding currently only applies to the right and bottom edges.

        Args:
            size (Union[PosInt, Tuple[PosInt, PosInt]]): Size (h, w) of the
                windows.
            stride (Union[PosInt, Tuple[PosInt, PosInt]]): Step size between
                windows. Can be 2-tuple (h_step, w_step) or positive int.
            padding (Optional[Union[PosInt, Tuple[PosInt, PosInt]]], optional):
                Optional padding to accomodate windows that overflow the
                extent. Can be 2-tuple (h_pad, w_pad) or non-negative int.
                If None, will be set to (size[0]//2, size[1]//2).
                Defaults to None.
            pad_direction (Literal['both', 'start', 'end']): If 'end', only pad
                ymax and xmax (bottom and right). If 'start', only pad ymin and
                xmin (top and left). If 'both', pad all sides. Has no effect if
                paddiong is zero. Defaults to 'end'.

        Returns:
            List[Box]: List of Box objects.
        """
        if not isinstance(size, tuple):
            size = (size, size)

        if not isinstance(stride, tuple):
            stride = (stride, stride)

        if size[0] <= 0 or size[1] <= 0 or stride[0] <= 0 or stride[1] <= 0:
            raise ValueError('size and stride must be positive.')

        if padding is None:
            padding = (size[0] // 2, size[1] // 2)

        if not isinstance(padding, tuple):
            padding = (padding, padding)

        if padding[0] < 0 or padding[1] < 0:
            raise ValueError('padding must be non-negative.')

        if padding != (0, 0):
            h_pad, w_pad = padding
            if pad_direction == 'both':
                padded_box = self.pad(
                    ymin=h_pad, xmin=w_pad, ymax=h_pad, xmax=w_pad)
            elif pad_direction == 'end':
                padded_box = self.pad(ymin=0, xmin=0, ymax=h_pad, xmax=w_pad)
            elif pad_direction == 'start':
                padded_box = self.pad(ymin=h_pad, xmin=w_pad, ymax=0, xmax=0)
            else:
                raise ValueError('pad_directions must be one of: '
                                 '"both", "start", "end".')
            return padded_box.get_windows(
                size=size, stride=stride, padding=(0, 0))

        # padding is necessarily (0, 0) at this point, so we ignore it
        h, w = size
        h_step, w_step = stride
        # lb = lower bound, ub = upper bound
        ymin_lb = self.ymin
        xmin_lb = self.xmin
        ymin_ub = self.ymax - h
        xmin_ub = self.xmax - w

        windows = []
        for ymin in range(ymin_lb, ymin_ub + 1, h_step):
            for xmin in range(xmin_lb, xmin_ub + 1, w_step):
                windows.append(Box(ymin, xmin, ymin + h, xmin + w))
        return windows

    def to_dict(self) -> Dict[str, int]:
        return {
            'xmin': self.xmin,
            'ymin': self.ymin,
            'xmax': self.xmax,
            'ymax': self.ymax
        }

    @classmethod
    def from_dict(cls, d: dict) -> 'Box':
        return cls(d['ymin'], d['xmin'], d['ymax'], d['xmax'])

    @staticmethod
    def filter_by_aoi(windows: List['Box'],
                      aoi_polygons: List[Polygon],
                      within: bool = True) -> List['Box']:
        """Filters windows by a list of AOI polygons

        Args:
            within: if True, windows are only kept if they lie fully within an
                AOI polygon. Otherwise, windows are kept if they intersect an AOI
                polygon.
        """
        result = []
        for window in windows:
            w = window.to_shapely()
            for polygon in aoi_polygons:
                if ((within and w.within(polygon))
                        or ((not within) and w.intersects(polygon))):
                    result.append(window)
                    break

        return result

    @staticmethod
    def within_aoi(window: 'Box', aoi_polygons: List[Polygon]) -> bool:
        """Check if window is within a list of AOI polygons."""
        w = window.to_shapely()
        for polygon in aoi_polygons:
            if w.within(polygon):
                return True
        return False



from rastervision.pipeline.config import Config, register_config


@register_config('raster_transformer')
class RasterTransformerConfig(Config):
    """Configure a :class:`.RasterTransformer`."""

    def update(self, pipeline=None, scene=None):
        pass

    def update_root(self, root_dir):
        pass



import numpy as np

from rastervision.core.data.raster_transformer.raster_transformer \
    import RasterTransformer


class NanTransformer(RasterTransformer):
    """Removes NaN values from float raster."""

    def __init__(self, to_value: float = 0.0):
        """Construct a new NanTransformer.

        Args:
            to_value: (float) NaN values are replaced
                with this
        """
        self.to_value = to_value

    def transform(self, chip, channel_order=None):
        """Transform a chip.

        Removes NaN values.

        Args:
            chip: ndarray of shape [height, width, channels] This is assumed to already
                have the channel_order applied to it if channel_order is set. In other
                words, channels should be equal to len(channel_order).

        Returns:
            [height, width, channels] numpy array

        """
        chip[np.isnan(chip)] = self.to_value
        return chip
